from bridges.data_src_dependent.data_source import *
from bridges.array import *
from bridges.avl_tree_element import *
from bridges.bst_element import *
from bridges.bridges import *
from bridges.bst_element import *
from bridges.circ_dl_element import *
from bridges.circ_sl_element import *
from bridges.color import *
from bridges.connector import *
from bridges.dl_element import *
from bridges.edge import *
from bridges.element import *
from bridges.element_visualizer import *
from bridges.graph_adj_list import *
from bridges.graph_adj_matrix import *
from bridges.grid import *
from bridges.link_visualizer import *
from bridges.ml_element import *
from bridges.sl_element import *
from bridges.tree_element import *
from bridges.color_grid import *
# games
from bridges.socket_connection import *
from bridges.keypress_listener import *
from bridges.gamebase import *
from bridges.non_blocking_game import *
from bridges.input_helper import *
from bridges.blocking_game import *
# line chart
from bridges.line_chart import *
# audio
from bridges.audio_channel import *
from bridges.audio_clip import *
# symbol collection
from bridges.symbol_collection import *
from bridges.symbol import *
from bridges.circle import *
from bridges.rectangle import *
from bridges.polygon import *
from bridges.polyline import *
from bridges.label import *






class LineChart:
    """
    @brief Show series of data or functions using a line chart.
    
    Line charts (https://en.wikipedia.org/wiki/Line_chart) are used to
    represent graphically functions such as f(x) = 3*x+1, or data such
    as temperature of a liquid on a stove as time passes. A individual
    function or a set of data is called "series".
    
    A series is represented by two arrays xdata and ydata such that
    there is a point at (xdata[0], ydata[0]), an other at (xdata[1],
    ydata[1]), ... One can add a series by passing the two arrays using
    setDataSeries() or add the arrays individually using setXData() and
    setYData().
    
    The different series have a label associated with them by default
    which can be disabled (see toggleSeriesLabel()).
    
    The data is typically shown with axes that use a linear
    scale. However, the scale can be changed to logarithmic for each
    axis individually (see toggleLogarithmicX() and
    toggleLogarithmic()).
    
    The LineChart can have a title (see getTitle() and setTitle()) and
    a subtitle (see setSubTitle() and getSubTitle()).
    
    @author Erik Saule, Matthew Mcquaigue
    
    @date 7/23/19
    """

    def __init__(self):
        self._plot_title = ""
        self._plot_subtitle= ""
        self._y_label = ""
        self._x_label = ""
        self.yaxis_data = dict()
        self.xaxis_data = dict()
        self._mouse_track = False
        self._data_label = True
        self._logarithmicx = False
        self._logarithmicy = False

    def get_data_structure_type(self):
        """
        Get the data type
        Returns: 
             name of the data type (used internally)
        """
        return "LineChart"

    @property
    def mouse_track(self) -> bool:
        """
        Is mounse tracking on? 
        Returns:
            bool: mouse tracking flag
        """
        return self._mouse_track

    @mouse_track.setter
    def mouse_track(self, val: bool):
        """
        Set mouse tracking flag
        Args:
           val(bool): mouse tracking flag
        """
        self._mouse_track = val

    @property
    def data_label(self):
        """
        Getter for data label
        Returns:
            bool: data label flag
        """
        return self._data_label

    @data_label.setter
    def data_label(self, val):
        """
        Setter for data label flag
        Args:
            val (bool) : data label flag 
        """
        self._data_label = val

    @property
    def logarithmicx(self):
        """
        use logarithmic scale on X axis?
        Returns:
            bool : logarthmic scale flag 
        """
        return self._logarithmicx

    @logarithmicx.setter
    def logarithmicx(self, val):
        """
        Setter for logarithmic scale on X axis
        Args:
            val (bool) : logarithmic scale flag
        Returns:
            None
        """
        self._logarithmicx = val

    @property
    def logarithmicy(self):
        """
        use logarithmic scale on Y axis?
        Returns:
            bool : logarthmic scale flag 
        """
        return self._logarithmicy

    @logarithmicy.setter
    def logarithmicy(self, val):
        """
        Setter for logarithmic scale on Y axis
        Args:
            val (bool) : logarithmic scale flag
        Returns:
            None
        """
        self._logarithmicy = val

    @property
    def title(self):
        """
        Getter for plot title
        Returns:
            str : plot title
        """
        return self._plot_title

    @title.setter
    def title(self, t):
        """
        Setter for plot title
        Args:
            t (str): plot title
        Returns:
            None
        """
        self._plot_title = t

    @property
    def sub_title(self):
        """
        Getter for plot sub title
        Returns:
            str : plot sub title
        """
        return self._plot_subtitle

    @sub_title.setter
    def sub_title(self, s):
        """
        Setter for plot sub title
        Args:
            s (str): plot sub title
        Returns:
            None
        """
        self._plot_subtitle = s

    @property
    def y_label(self):
        """
        Getter for plot Y label
        Returns:
            str : plot label
        """
        return self._y_label

    @y_label.setter
    def y_label(self, label):
        """
        Setter for plot Y label
        Args:
            label (str): plot label
        Returns:
            None
        """
        self._y_label = label

    @property
    def x_label(self):
        """
        Getter for plot X label
        Returns:
            str : plot label
        """
        return self._x_label

    @x_label.setter
    def x_label(self, label):
        """
        Setter for plot X label
        Args:
            label (str): plot label
        Returns:
            None
        """
        self._x_label = label

    def set_data_series(self, series_name, x_data, y_data):
        """
        Setter for plot data on X and Y axes
        Args:
            series_name: title for the series
            x_data:  plot data for x axis
            y_data:  plot data for y axis
        Returns:
            None
        """
        self.set_x_data(series_name, x_data)
        self.set_y_data(series_name, y_data)

    def set_x_data(self, series, x_data):
        """
        Setter for plot data on X axis
        Args:
            series: the name of the series
            x_data:  plot data
        Returns:
            None
        """
        self.xaxis_data[series] = x_data

    def get_x_data(self, series):
        """
        Getter for plot data on X
        Returns:
            dict : plot data
        """
        return self.xaxis_data[series]

    def set_y_data(self, series, y_data):
        """
        Setter for plot data on Y axis
        Args:
            series:  series name
            y_data:  plot data
        """
        self.yaxis_data[series] = y_data

    def get_y_data(self, series):
        """
        Getter for plot data on Y
        Returns:
            dict : plot data
        """
        return self.yaxis_data[series]

    def convert(self, x_data):
        """
        Convert input data into an array of floats.
        Args:
            x_data : input data in a list of values
        Returns:
            A list of floats of the same data
        """
        arr = []
        for i in range(len(x_data)):
            arr.append(float(x_data[i]))
        return arr

    def check(self) -> bool:
        correct = True
        for key, value in self.xaxis_data.items():
            series = key
            xdata = value
            ydata = self.yaxis_data.get(series)
            if ydata is None:
                print("Series \"" + series + "\" has xdata but no ydata")
                correct = False
            if len(xdata) != len(ydata):
                print("Series \"" + series + "\" has xdata and ydata of different sizes")
                correct = False
            if self.logarithmicx:
                for i in range(0, len(xdata)):
                    if xdata[i] <= 0:
                        print("Xaxis scale is logarithmic but series " + series +
                              " has xdata[" + str(i) + "] = " + str(xdata[i]) + " (should be stricly positive)")
            if self.logarithmicy:
                for i in range(0, len(ydata)):
                    if ydata[i] <= 0:
                        print("Yaxis scale is logarithmic but series " + series +
                              " has ydata[" + str(i) + "] = " + str(ydata[i]) + " (should be stricly positive)")
        for key, value in self.yaxis_data.items():
            series = key
            ydata = value
            xdata = self.xaxis_data.get(series)
            if xdata is None:
                print("Series: " + series + " has ydata but no xdata")
                correct = False

        return correct

    def get_data_structure_representation(self):
        """
        Get a JSON of the data structure representation
        Returns:
            JSON string of the data structure
        """
        self.check()

        xaxis_json = []
        yaxis_json = []
        for key, value in self.xaxis_data.items():
            x_temp = []
            data_key = key
            data_value = value
            for i in range(0, len(data_value)):
                x_temp.append(data_value[i])
            temp_xaxis_json = {
                "Plot_Name": str(data_key),
                "xaxis_data": x_temp
            }
            xaxis_json.append(temp_xaxis_json)


        for key, value in self.yaxis_data.items():
            y_temp = []
            data_key = key
            data_value = value
            for i in range(0, len(data_value)):
                y_temp.append(data_value[i])
            temp_yaxis_json = {
                "Plot_Name": str(data_key),
                "yaxis_data": y_temp
            }
            yaxis_json.append(temp_yaxis_json)


        json_dict = {
            "plot_title": str(self.title),
            "subtitle": str(self.sub_title),
            "xLabel": str(self.x_label),
            "yLabel": str(self.y_label),
            "xaxisType": self.logarithmicx,
            "yaxisType": self.logarithmicy,
            "options": {
                "mouseTracking": self._mouse_track,
                "dataLabels": str(self.data_label)
            },
            "xaxis_data": xaxis_json,
            "yaxis_data": yaxis_json
        }

        return json_dict



class GutenbergMeta:
    """
    @brief Class to hold the meta data of various books from the Gutenberg Project. 

    This is to be used with the Gutenberg request functions. 

    This class holds id, title, language, date_added,
    authors, genres, and library of congress classifications.

    This object is generally not created by the user, to see how its created check 
    out bridges::data_src_dependent::data_source::get_gutenberg_book_metadata()

    @sa A tutorial of how to use the Gutenberg data in BRIDGES is 
         available: check out https://bridgesuncc.github.io/tutorials/Data_Gutenberg.html
    @author Jay Strahler
    @date 12/28/20 
    """

    def __init__(self, id = 0, title = "", lang = "", date_added = "", authors = [], genres = [], loc = []):
        """
        Constructor for a gutenbergmeta object that holds the metadata
        for a gutenberg book
        :param id: book id
        :param title: book title
        :param lang: book language
        :param date_added: data the book was added
        :param authors: authors of the book
        :param genres: book genres
        :param loc:
        """
        self.id = id
        self.title = title
        self.lang = lang
        self.date_added = date_added
        self.authors_list = authors
        self.genres_list = genres
        self.loc_list = loc

    @property
    def id(self):
        """
        @brief get book's id
        Returns:
            id of book
        """
        return self.__id

    @id.setter
    def id(self, i):
        """
        @brief Set book's id
        Args:
            i:  id of book
        """
        self.__id = i

    @property
    def title(self):
        """
        @brief get book's title
        Returns:
            title of book
        """
        return self.__title

    @title.setter
    def title(self, i):
        """
        @brief Set book's title
        Args:
            i:  title of book
        """
        self.__title = i

    @property
    def lang(self):
        """
        @brief get book's language
        Returns:
            language of book
        """
        return self.__lang
    
    @lang.setter
    def lang(self, i):
        """
        @brief Set book's language
        Args:
            i:  language of book
        """
        self.__lang = i

    @property
    def date(self):
        """
        @brief get book's date added to gutenberg
        Returns:
            date of book
        """
        return self._date_added
    
    @date.setter
    def date(self, i):
        """
        @brief Set book's date added to gutenberg
        Args:
            i:  date of book
        """
        self.__date_added = i

    @property
    def authors(self):
        """
        @brief get book's authors
        Returns:
            authors of book
        """
        return self.__authors_list

    @authors.setter
    def authors(self, i):
        """
        @brief Set book's authors
        Args:
            i:  authors of book
        """
        self.__authors_list = i

    @property
    def genres(self):
        """
        @brief get book's genres
        Returns:
            genres of book
        """
        return self.__genres_list

    @genres.setter
    def genres(self, i):
        """
        @brief Set book's genres
        Args:
            i:  genres of book
        """
        self.__genres_list = i

    @property
    def loc(self):
        """
        @brief get book's library of congress classifications
        Returns:
            library of congress calssifications of book
        """
        return self.__loc_list

    @loc.setter
    def loc(self, i):
        """
        @brief Set book's library of congress classifications
        Args:
            i:  library of congress classifications of book
        """
        self.__loc_list = i



class GutenbergBook:
    """
    @brief  A Gutenberg Book object metadata only, used along with the books data source.
 
    This is a convenience class provided for  users who wish to use this
    data source as part of their application. It provides an API that makes
    it easy to access the attributes of this data set.
   
    Refer to tutorial examples to using this data source in data structure
    assignments.
     
    @author Matthew Mcquaigue, Kalpathi Subramanian
    @date   2/1/17, 12/29/20 
    """

    def __init__(self, author_name: str = "", author_birth: int = 0, author_death: int = 0, title: str = "", lang: str = "", genre: str = "", subject: str = "", num_chars: int = 0, num_words: int = 0, num_sentences: int = 0, num_difficult_words: int = 0, url: str = "", downloads: int = 0):
        """ 
        @brief constructor
        Args:
            author_name: name of author [string]
            author_birth: author birth date [int]
            author_death: aurthor death date [int]
            title:  title of book [string]
            lang:   language of book [string]
            genre:  genres of book [list of strings]
            subject: subject of book [string]
            num_chars: number of characters [int]
            num_words: number of words [int]
            num_sentences: number of sentences [int]
            num_difficult_words: number of difficult words [int]
            url:  url of book [string]
            downloads: number of downloads of book [int]
        """
        self._author_name = author_name
        self._author_birth = author_birth
        self.author_death = author_death
        self._title = title
        self._lang = lang
        self._genre = genre
        self._subject = subject
        self._num_chars = num_chars
        self._num_words = num_words
        self._num_sentences = num_sentences
        self._num_difficult_words = num_difficult_words
        self._url = url
        self._downloads = downloads

    @property
    def name(self):
        """
        @brief get author's name
        Returns:
            name of author
        """
        return self._author_name

    @name.setter
    def name(self, n):
        """
        @brief Set author's name
        Args:
            n:  name of author
        """
        self._author_name = n

    @property
    def birth(self):
        """
        @brief get birth date of book's author
        Returns:
            birthdate of author
        """
        return self._author_birth

    @birth.setter
    def birth(self, b):
        """
        @brief Set author's birth date
        Args:
            b:  author birth date to set
        """
        self._author_birth = b

    @property
    def death(self):
        """
        @brief get date of death of book's author
        Returns:
            date of death of author
        """
        return self._author_death

    @death.setter
    def death(self, d):
        """
        @brief Set date of death of book's author
        Args:
            d: author's date of death to be set
        """
        self._author_death = d

    @property
    def title(self):
        """
        @brief Get title of book
        Returns:
            book title
        """
        return self._title

    @title.setter
    def title(self, t):
        """
        @brief Set book title
        Args:
            t: title to set
        """
        self._title = t

        """
        @brief Get language of book
        Returns:
            book language
        """
    @property
    def lang(self):
        """
        @brief Get book's language
        Returns:
            book's language
        """
        return self._lang

    @lang.setter
    def lang(self, l):
        """
        @brief Set book's language
        Args:
            l: language to set
        """
        self._lang = l

    @property
    def genre(self):
        """
        @brief Get book's genres
        Returns:
            book's genres
        """
        return self._genre

    @genre.setter
    def genre(self, g):
        """
        @brief Set book's genres
        Args:
            g: genres to set
        """
        self._genre = g

    @property
    def subject(self):
        """
        @brief Get book's subject
        Returns:
            book's subject
        """
        return self._subject

    @subject.setter
    def subject(self, s):
        """
        @brief Set book's subject
        Args:
            s: subject to set
        """
        self._subject = s

    @property
    def url(self):
        """
        @brief Get book's url
        Returns:
            book's url
        """
        return self._url

    @url.setter
    def url(self, u):
        """
        @brief Set book's url
        Args:
            u: url to set
        """
        self._url = u

    @property
    def num_chars(self):
        """
        @brief Get number of characters in book
        Returns:
            book's character count
        """
        return self._num_chars

    @num_chars.setter
    def num_chars(self, n):
        """
        @brief Set book's number of characters
        Args:
            n: number of characters to set
        """
        self._num_chars = n

    @property
    def num_words(self):
        """
        @brief Get number of words in book
        Returns:
            book's words count
        """
        return self._num_words

    @num_words.setter
    def num_words(self, w):
        """
        @brief Set book's number of words
        Args:
            w: number of words to set
        """
        self._num_words = w

    @property
    def num_sentences(self):
        """
        @brief Get number of sentences in book
        Returns:
            book's sentence count
        """
        return self._num_sentences

    @num_sentences.setter
    def num_sentences(self, n):
        """
        @brief Set book's number of sentences
        Args:
            n: number of sentences to set
        """
        self._num_sentences = n

    @property
    def num_difficult_words(self):
        """
        @brief Get number of difficulut words in book
        Returns:
            book's difficulut word count
        """
        return self._num_difficult_words

    @num_difficult_words.setter
    def num_difficult_words(self, w):
        """
        @brief Set book's number of difficult words
        Args:
            w: difficult word count to set
        """
        self._num_difficult_words = w

    @property
    def downloads(self):
        """
        @brief Get number of downloads of book
        Returns:
            book's download count
        """
        return self._downloads

    @downloads.setter
    def downloads(self, d):
        """
        @brief Set book's number of downloads
        Args:
            d: download count to set
        """
        self._downloads = d




from bridges.bst_element import *


class KDTreeElement(BSTElement):
    """
    @brief The This class can be used to create K-d Tree  elements, derived
    from BSTElement. K-D trees can be thought of as the spatial equivalent
    binary search trees and operate on multiple dimensions (2D and 3D are
    most common). These trees serve as a representation of the underlying
    geometrically defined spaces. Specialized versions of these trees
    include quadtrees and octrees, which subdivide into equal sized quadrants or
    octants at each level, respectively.
    This class extends the BSTElement class by adding a dimension property.
    It also includes a thickness property for displaying the partitioning
    lines generated by the convex decomposition.
    
    @author Kalpathi Subramanian, Matthew McQuaigue
    @date 12/26/18, 7/23/19
    
    \sa Kd Tree tutorial, https://bridgesuncc.github.io/tutorials/KdTree.html
    """
    def __init__(self, **kwargs):
        """
        Constructor for a kdTreeElement with the provided value, label, key, left and right kdTree elements.
        The defaults will be used if not provided
        Args:
            (object) key: The key for the ordering
            val: the data this element will hold
            (str) label: the label to show during visualization
            left: the left kdTree element
            right:  the right kdTree element
            (float) th: thickness of KDTree partition lines
            (int) dim: partition dimension
        Returns:
            None
        """
        if 'val' in kwargs:
            if 'label' in kwargs:
                if 'left' and 'right' in kwargs:
                    super(KDTreeElement, self).__init__(e = kwargs['val'], label = kwargs['label'],
                                                        left = kwargs['left'], right = kwargs['right'])
                else:
                    super(KDTreeElement, self).__init__(e=kwargs['val'], label=kwargs['label'])
            else:
                if 'left' and 'right' in kwargs:
                    super(KDTreeElement, self).__init__(e=kwargs['val'], left=kwargs['left'], right=kwargs['right'])
                else:
                    super(KDTreeElement, self).__init__(e=kwargs['val'])
        else:
            super(KDTreeElement, self).__init__()
        if 'key' in kwargs:
            super(KDTreeElement, self.__class__).key.fset(self, kwargs['key'])
        if 'th' in kwargs:
            self._thickness = kwargs['th']
        else:
            self._thickness = 0.0
        if 'dim' in kwargs:
            self._dimension = kwargs['dim']
        else:
            self._dimension = 0

    def get_data_structure_type(self) -> str:
        """
        Getter for the data structure type
        Returns:
            str: of the the data structure type
        """
        return "KdTree"

    @property
    def dimension(self) -> int:
        """
        Getter for the dimensions of the partition lines
        Returns:
            int: dimension (0, 1, 2, etc)
        """
        return self._dimension

    @dimension.setter
    def dimension(self, dim: int) -> None:
        """
        Setter for the dimension of th partitioning at this tree node
        Args:
            (int) dim: dimension value to set
        Returns:
            None
        """
        self._dimension = dim

    @property
    def thickness(self) -> float:
        """
        Getter for the thickness to the KDTree partitions
        Returns:
            float: the thickness
        """
        return self._thickness

    @thickness.setter
    def thickness(self, th) -> None:
        """
        Setter for the thickness of the KDTree partitions - can be used in the visualization
        Args:
            (float) th: thickness of partitions
        Returns:
            None
        """
        self._thickness = th

    @property
    def left(self):
        """
        Getter for the left child of the tree element
        Returns:
            KdTreeElement: the left child
        """
        return super(KDTreeElement, self).left

    @left.setter
    def left(self, l):
        """
        Set left child
        Args:
            l: left child to set
        Returns:
            None
        """
        self.set_child(0, l)

    @property
    def right(self):
        """
        Getter for the right child of the tree element
        Returns:
            KdTreeElement: the right child
        """
        return super(KDTreeElement, self).right

    @right.setter
    def right(self, r):
        """
        Set right child
        Args:
            r: right child to set
        Returns:
            None
        """
        self.set_child(1, r)

    def get_element_representation(self):
        orig_json_str = super(KDTreeElement, self).get_element_representation()
        kdt_dict = {
            "dimension": str(self.dimension),
            "thickness": str(self.thickness)
        }
        orig_json_str.update(kdt_dict)
        return orig_json_str






class Camera:

    def __init__(self, type):
        self._type = type
        self._object_json = {
            'type': type
        }

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type):
        self._type = type

    def push_representation(self, scene_json):
        scene_json['camera'] = self._object_json





from bridges.array import *

class Array3D(Array):
    """
    @brief This class can be used to create three dimensional arrays of 
    type Element<E>.
    
    @author 	Matthew Mcquaigue, Kalpathi Subramanian
    
    @date  	7/18/19, 12/29, 20
    
    This class can be used to create three dimensional arrays of 
    type Element<E>  where E is a generic object representing application 
    specific data.
    
    Arrays are internally represented as 1D arrays; currently 1D, 2D  and
    3D arrays are supported.
    
    Example Tutorial at: https://bridgesuncc.github.io/tutorials/Array.html 
    (1D, 2D, and 3D Array)<br>
    """

    def __init__(self, **kwargs):
        """
        Create an array object with the specified dimensions
        Kwargs:
            (list) dims: size of each dimension
            (int) rows: number of rows
            (int) cols: number of columns
            (int) slices: number of slices
        """
        super(Array3D, self).__init__()
        if 'dims' in kwargs:
            self.set_size(3, kwargs['dims'])
            self.num_cols = kwargs['dims'][0]
            self.num_rows = kwargs['dims'][1]
            self.num_slices = kwargs['dims'][2]
        elif 'slices' in kwargs and 'rows' in kwargs and 'cols' in kwargs:
            dims = [kwargs['cols'], kwargs['rows'], kwargs['slices']]
            self.set_size(3, dims)
            self.num_cols = kwargs['cols']
            self.num_rows = kwargs['rows']
            self.num_slices = kwargs['slices']
        else:
            self.num_cols = 0
            self.num_rows = 0
            self.num_slices = 0
            self.size = 0



from bridges.line_chart import *
from bridges.graph_benchmark import *
from bridges.data_src_dependent.data_source import *
from datetime import datetime
import time as time_

class ShortestPathBenchmark(GraphBenchmark):

    def __init__(self, p):
        super(ShortestPathBenchmark,self).__init__()
        self.__plot = p
        self.__plot.x_label = "Number of Edges"
        self.__plot.y_label = "Runtime (in s)"

    def __get_center(self, osm_data, graph, latc, lonc):
        def dist_function(v):
            return (v.latitude - latc) * (v.latitude - latc) + (v.logitude - lonc) * \
            (v.logitude - lonc)

        minindex = 0
        dist = dist_function(graph.get_vertex(minindex).value)

        for i in range(1, len(graph.vertices)):
            locdist = dist_function(graph.get_vertex(i).value)
            if locdist < dist:
                minindex = i
                dist = locdist

        return minindex

    def run(self, algoname, spalgo):
        time = []
        vtx_count = []
        edge_count = []

        reflat = 40.74
        reflong = -73.98

        radius = 0.2
        while radius < 0.15:
            osm_data = get_osm_data(reflat - radius, reflong - radius, reflat+radius, reflong+radius)
            graph = osm_data.get_graph()
            vtx_count = self._count_vertices(graph)
            edge_count = self._count_edges(graph)
            root = self.__get_center(osm_data, graph, reflat, reflong)

            level = dict()
            parent = dict()

            start = int(round(time_.time() * 1000))
            spalgo(graph, root, level, parent)
            end = int(round(time_.time() * 1000))

            elapsed_seconds = end - start
            time.append(elapsed_seconds)
            vtx_count.append(vtx_count)
            edge_count.append(edge_count)

            if elapsed_seconds > self.time_cap:
                break
        self.__plot.set_x_data(algoname, edge_count)
        self.__plot.set_y_data(algoname, time)



import datetime
import time

class EarthquakeUSGS:
    """
    @brief Class that holds earthquake data records.

    Class that hold earthquake data, for use with USGIS retrieved quake data.
    BRIDGES uses scripts to 
    continually monitor USGIS site (tweets) and retrieve the latest 
    quake data for use in student projects.

    This object is generally not created by the user, to see how its created check 
    out bridges::data_src_dependent::data_source::get_earthquake_usgs_data()

    @sa For an example, check out https://bridgesuncc.github.io/tutorials/Data_EQ_USGS.html

    @author Matthew Mcquaigue, Kalpathi Subramanian,
    @date  2/18/18, 12/29/20, 1/6/21
    """
    def __set_time_from_unix_timestamp(self, tm):
        epoch_time = int(tm)
        eq_time = epoch_time / 1000
        eqt = time.gmtime(eq_time)
        self._time = time.strftime("%Y-%m-%d %H:%M:%S", eqt)

    def __init__(self, magnitude=None, longit=None, latit=None, location=None, 
                     title=None, url=None, time=None):
        """
        @brief constructor
        Args: 
            magnitude: magnitude of quake 
            latit: latitude position
            longit: longitude position
            location:  location of quake
            title:     title (has some of eq info in a string)
            url:       url for more information
            time:       occurrence time of quake
        """
        self._time = int()
        if magnitude is not None:
            self._magnitude = magnitude
        else:
            self._magnitude = 0.0
        if longit is not None:
            self._longit = longit
        else:
            self._longit = 0.0
        if latit is not None:
            self._latit = latit
        else:
            self._latit = 0.0
        if location is not None:
            self._location = location
        else:
            self._location = ""
        if title is not None:
            self._title = title
        else:
            self._title = ""
        if url is not None:
            self._url = url
        else:
            self._url = ""
        if time is not None:
            self.time =  time

    @property
    def time(self):
        """   
        @brief Get occurrence time (epoch)  of quake
        Returns:
            Quake occurrence time
        """
        return self._time

    @time.setter
    def time(self, tm) -> None:
        """   
        @brief Set occurrence time (epoch)  of quake
        Args:
            tm: Quake occurrence time to set
        """
        self.__set_time_from_unix_timestamp(tm)

    @property
    def latit(self) -> float:
        """   
        @brief Get latitude of quake
        Returns:
            Quake latitude
        """
        return self._latit

    @latit.setter
    def latit(self, latit: float) -> None:
        """   
        @brief Set latitude of quake
        Args:
            latit: quake latitude to set
        """
        self._latit = latit

    @property
    def longit(self) -> float:
        """   
        @brief Get longitude of quake
        Returns:
            Quake longitude
        """
        return self._longit

    @longit.setter
    def longit(self, longit: float) -> None:
        """   
        @brief Set longitude of quake
        Args:
            longit: quake longitude to set
        """
        self._longit = longit

    @property
    def location(self) -> str:
        """   
        @brief Get location of quake (typically a city or something of the sort)
        Returns:
            Quake location 
        """
        return self._location

    @location.setter
    def location(self, location: str):
        """   
        @brief Set location of quake
        Args:
            location: quake location to set
        """
        self._location = location

    @property
    def title(self) -> str:
        """   
        @brief Get quake title 
        Returns:
            Quake title 
        """
        return self._title

    @title.setter
    def title(self, title: str):
        """   
        @brief Set title of quake
        Args:
            title: quake title to set
        """
        self._title = title

    @property
    def url(self) -> str:
        """   
        @brief Get quake url 
        Returns:
            Quake url 
        """
        return self._url

    @url.setter
    def url(self, url: str):
        """   
        @brief Set url of quake
        Args:
            url: quake url to set
        """
        self._url = url

    @property
    def magnitude(self) -> float:
        """   
        @brief Get quake magnitude (Richter scale) 
        Returns:
            Quake magnitude 
        """
        return self._magnitude

    @magnitude.setter
    def magnitude(self, magn: float):
        """
        Setter for the magnitude of the quake
        Args:
            magn: magnitude to set
        """
        self._magnitude = magnitude



class Game:
    """
    @brief  A Game object, used along with the Games data source.
    
    This is a convenience class provided for  users who wish to use this
    data source as part of their application. It provides an API that makes
    it easy to access the attributes of this data set.
    
    Each game record has title, platform on which it can be played,
    rating,  and a list of genres.

    This object is generally not created by the user, to see how its created check 
    out bridges::data_src_dependent::data_source::get_game_data()

    @sa For an example, check out https://bridgesuncc.github.io/tutorials/Data_IGN_Games.html
    
    @author Matthew Mcquaigue, Kalpathi Subramanian
    @date   2/1/17, 12/29/20, 1/6/21
    """

    def __init__(self, title: str = "", platform: str = "", rating: float = 0.0, genre: str = ""):
        """
        @brief Constructor
        Args:
            title:  game title
            platform: game platform
            rating:  game rating
            genre:   game's genres
        """
        self._title = title
        self._platform = platform
        self._rating = rating
        self._genre = genre

    @property
    def title(self):
        """
        @brief get game title
        Returns:
            game title
        """
        return self._title

    @title.setter
    def title(self, t):
        """
        @brief Set game title
        Args:
            t: game title to set
        """
        self._title = t

    @property
    def platform(self):
        """
        @brief get game platform
        Returns:
            game platform
        """
        return self._platform

    @platform.setter
    def platform(self, p):
        """
        @brief Set game platform
        Args:
            p: game platform to set
        """
        self._platform = p

    @property
    def rating(self):
        """
        @brief get game rating
        Returns:
            game rating
        """
        return self._rating

    @rating.setter
    def rating(self, r):
        """
        @brief Set game rating
        Args:
            r: game rating to set
        """
        self._rating = r

    @property
    def genre(self):
        """
        @brief get game genres
        Returns:
            game genres (list of strings)
        """
        return self._genre

    @genre.setter
    def genre(self, g):
        """
        @brief Set game title
        Args:
            g:  game genres to set
        """
        self._genre = g





class Grid:
    """
    @brief This is a base class in BRIDGES for representing an (n x n) grid.
    
    @author David Burlinson, Matthew McQuaigue

    @date  2018, 7/24/19, 2021

    \sa Color grid tutorial at https://bridgesuncc.github.io/tutorials/Grid.html
    """
    import traceback

    grid_size = [10, 10]
    maxGridSize = [1080, 1920]

    def get_data_structure_type(self) -> str:
        """
        Get the data structure type
        Returns:
            str: representing the data structure type
        """
        return "Grid"

    def __init__(self, **kwargs) -> None:
        """
        Grid constructor
        Kwargs:
            size: size of the grid as array
            rows: number of rows in grid
            cols: number of the columns in grid
        Returns:
            None
        Raises:
            ValueError: if the size dimensions are greater than the max grid sizes (1080, 1920)
        """
        if 'size' in kwargs:
            if (kwargs['size'][0] <= 0 or kwargs['size'][0] > Grid.maxGridSize[0]) or \
                    (kwargs['size'][1] <= 0 or kwargs['size'][1] > Grid.maxGridSize[1]):
                raise ValueError("Invalid size: [" + str(
                    str(kwargs['size'][0]) + "," + str(kwargs['size'][1]) + "] .. please use values between (0 and " +
                    str(self.maxGridSize[0]) + "] for rows and values between (0 and " + str(
                        self.maxGridSize[1]) + "] for columns"))
            self.grid = []
            for i in range(kwargs['size'][0]):
                self.grid.append([])
                for j in range(kwargs['size'][1]):
                    self.grid[i].append(None)
        if 'rows' in kwargs and 'cols' in kwargs:
            size = [kwargs['rows'], kwargs['cols']]
            if (size[0] <= 0 or size[0] > self.maxGridSize[0]) or (size[1] <= 0 or size[1] > self.maxGridSize[1]):
                raise ValueError("Invalid size: [" + str(
                    str(size[0]) + "," + str(size[1]) + "] .. please use values between (0 and " + str(
                        self.maxGridSize[0]) + "] for rows and values between (0 and " + str(
                        self.maxGridSize[1]) + "] for columns"))
            self.grid = []
            for i in range(size[0]):
                self.grid.append([])
                for j in range(size[1]):
                    self.grid[i].append(None)

    @property
    def dimensions(self) -> list:
        """
        Getter for the dimensions of the grid
        Returns:
            list: as the dimensions of the grid
        """
        return [self.grid_size[0], self.grid_size[1]]

    def get(self, row: int, col: int):
        """
        Get the row,col element in the grid
        Args:
            (int) row: row the element is in
            (int) col: col the element is in
        Returns:
            element, none
        Raises
            Exception: printing the traceback stack if returning the row and column is an error
        """
        try:
            return self.grid[row][col]
        except Exception as e:
            traceback.print_tb(e.__traceback__)
            return None

    def set(self, row: int, col: int, val) -> None:
        """
        set the (row, col) element in the grid
        Args:
            row: row position
            col: column position
            val: value to be set in (row, col) position
        Returns:
            None
        Raises:
            Exception: if setting the element at the row and column is a problem
        """
        try:
            self.grid[int(row)][int(col)] = val
        except Exception as e:
            traceback.print_tb(e.__traceback__)
            return None



import json
import requests
import pickle
import urllib.parse

from bridges.data_src_dependent import earthquake_usgs
from bridges.data_src_dependent import actor_movie_imdb
from bridges.data_src_dependent import game
from bridges.data_src_dependent import shakespeare
from bridges.data_src_dependent import gutenberg_book
from bridges.data_src_dependent import gutenberg_meta
from bridges.data_src_dependent import reddit
from bridges.data_src_dependent import cancer_incidence
from bridges.data_src_dependent import song
from bridges.data_src_dependent import lru_cache
from bridges.data_src_dependent import movie_actor_wiki_data
from bridges.data_src_dependent import us_cities
from bridges.data_src_dependent.osm import *
from bridges.data_src_dependent.elevation import *
from bridges.data_src_dependent.amenity import *
from bridges.data_src_dependent.actor_movie_imdb import *
from bridges.color_grid import ColorGrid
from bridges.color import Color
from SPARQLWrapper import SPARQLWrapper, JSON


source_type = "live"

def set_source_type(t):
    global source_type
    if (t == "testing"):
        source_type = "testing"
    elif (t == "local"):
        source_type = "local"
    return

def get_reddit_url():
    if source_type == "testing":
        return "http://bridges-data-server-reddit-t.bridgesuncc.org"
    elif source_type == "local":
        return "http://localhost:9999"
    else:
        return "http://bridges-data-server-reddit.bridgesuncc.org"

def get_gutenberg_url():
    if source_type == "testing":
        return "http://bridges-data-server-gutenberg-t.bridgesuncc.org"
    elif source_type == "local":
        return "http://localhost:3000"
    return "http://bridges-data-server-gutenberg.bridgesuncc.org"

def _get_osm_baseurl():
    if source_type == "local":
        return "http://localhost:3000"
    return "http://bridges-data-server-osm.bridgesuncc.org"

def get_elevation_url():
    if source_type == "local":
        return "http://localhost:3000"
    return "http://bridges-data-server-elevation.bridgesuncc.org"

def get_amenity_url():
    if source_type == "local":
        return "http://localhost:3000"
    return "http://bridges-data-server-osm.bridgesuncc.org"

def get_game_data():
    """
    @brief Get meta data of the IGN games collection.

    This function retrieves  and formats the data into a list of
    Game objects

    @throws Exception if the request fails
    
    @return a list of Game objects,
    """

    wrapper = []

    url = "http://bridgesdata.herokuapp.com/api/games"
    PARAMS = {"Accept: application/json"}

    r = requests.get(url=url, params=str(PARAMS))

    r = r.json()

    D = r["data"]
    # print(D)

    for i in range(len(D)):
        V = D[i]
        G = V["genre"]
        genre = []
        for j in range(len(G)):
            genre.append(str(G[j]))
        wrapper.append(game.Game(V["game"], V["platform"], V["rating"], str(genre)))
    return wrapper


def get_us_cities_data(**kwargs):
    wrapper = []

    url = "http://bridgesdata.herokuapp.com/api/us_cities"
    if len(kwargs) > 0:
        url = url + '?'
        if kwargs.get('state'):
            url = url + 'state=' + kwargs['state'] + '&'
        if kwargs.get('population'):
            url = url + 'population=' + str(kwargs['population']) + '&'
        if kwargs.get('minll'):
            url = url + 'minLatLong=' + str(kwargs['minll'][0]) + ',' + str(kwargs['minll'][1]) + '&'
        if kwargs.get('maxll'):
            url = url + 'maxLatLong=' + str(kwargs['maxll'][0]) + ',' + str(kwargs['maxll'][1]) + '&'
        if kwargs.get('elevation'):
            url = url + 'elevation=' + str(kwargs['elevation']) + '&'
        if kwargs.get('limit'):
            url = url + 'limit=' + str(kwargs['limit']) + '&'
        url = url[:-1]  # remove last &

    print(url)
    PARAMS = {"Accept: application/json"}

    r = requests.get(url=url, params=str(PARAMS))

    r = r.json()

    D = r["data"]

    for i in range(len(D)):
        V = D[i]
        wrapper.append(us_cities.USCities(city = V['city'], state=V['state'], country = V['country'], lat = V['lat'], lon=V['lon'], elevation=V['elevation'],
                                          population = V['population'], timezone=V['timezone']))
    return wrapper

def get_world_cities_data(**kwargs):
    wrapper = []

    url = "http://localhost:3001/api/world_cities"
    # url = "http://bridgesdata.herokuapp.com/api/world_cities"
    if len(kwargs) > 0:
        url = url + '?'
        if kwargs.get('state'):
            url = url + 'state=' + kwargs['state'] + '&'
        if kwargs.get('population'):
            url = url + 'population=' + str(kwargs['population']) + '&'
        if kwargs.get('minll'):
            url = url + 'minLatLong=' + str(kwargs['minll'][0]) + ',' + str(kwargs['minll'][1]) + '&'
        if kwargs.get('maxll'):
            url = url + 'maxLatLong=' + str(kwargs['maxll'][0]) + ',' + str(kwargs['maxll'][1]) + '&'
        if kwargs.get('elevation'):
            url = url + 'elevation=' + str(kwargs['elevation']) + '&'
        if kwargs.get('limit'):
            url = url + 'limit=' + str(kwargs['limit']) + '&'
        url = url[:-1]  # remove last &

    print(url)
    PARAMS = {"Accept: application/json"}

    r = requests.get(url=url, params=str(PARAMS))

    r = r.json()

    D = r["data"]

    for i in range(len(D)):
        V = D[i]
        wrapper.append(us_cities.USCities(city = V['city'], state=V['state'], country = V['country'], lat = V['lat'], lon=V['lon'], elevation=V['elevation'],
                                          population = V['population'], timezone=V['timezone']))
    return wrapper




def parse_actor_movie_imdb(item):
    """
    @brief Parses an actor movie pair
    Args:
        item: input actor-movie pair
    """
    am_pair = ActorMovieIMDB()
    am_pair.actor = item["actor"]
    am_pair.movie = item["movie"]
    return am_pair


def get_actor_movie_imdb_data(count = 0):
    """
    @brief Get ActorMovie IMDB Data. 

    Data is retrieved, formatted into a list of ActorMovieIMDB objects

    Args:
        count: count of the number of actor/movie pairs, but currently unused,
        returns all records.

    Returns:
        a list of ActorMovieIMDB objects, but only actor and movie fields in this version
    @throws Exception if the request fails
    """

    wrapper = []

    url = "http://bridgesdata.herokuapp.com/api/imdb?limit=" + str(count)
    PARAMS = {"Accept: application/json"}

    r = requests.get(url=url, params=str(PARAMS))
    # print(r.status_code, r.reason)

    data = r.json()

    D = data["data"]

    for i in range(len(D)):
        V = D[i]
        wrapper.append(actor_movie_imdb.ActorMovieIMDB(V["actor"], V["movie"]))

    return wrapper

def get_actor_movie_imdb_data2():
    """
    Get ActorMovie IMDB Data. Data is retrieved, formatted into a list of ActorMovieIMDB objects
    Args:
        count: the number of actor/movie pairs, but currently unused,
    Returns:
        All records.
    Throws: 
        Exception if the request fails
    """
    url = "https://bridgesdata.herokuapp.com/api/imdb2"

    r = requests.get(url = url)
    status = r.status_code

    data = r.json()
    D = data["data"]
    am_list = []

    if status == 200:
        for i in range(len(D)):
            V = D[i]
            am_pair = parse_actor_movie_imdb(V)
            am_pair.rating = int(V['rating'])

            genre = V['genres']
            v = []
            for k in range(len(genre)):
                v.append(genre[k])
            am_pair._genres = v
            am_list.append(am_pair)
        return am_list
    else:
        r.raise_for_status()


def get_earthquake_usgs_data(count = 0):
    """
    @brief Get USGS earthquake data

    USGS Tweet data (https://earthquake.usgs.gov/earthquakes/map/). Data
    retrieved, formatted into a list of EarthquakeUSGS objects

    Args:
        count: the number of earthquake records retrieved

    @throws Exception if the request fails

    Returns:
        a list of earthquake records
    """
    wrapper = []
    url = "http://earthquakes-uncc.herokuapp.com/eq"
    latest_url = "http://earthquakes-uncc.herokuapp.com/eq/latest/" + str(count)
    PARAMS = {"Accept: application/json"}

    if count <= 0:
        r = requests.get(url=url, params=str(PARAMS))
        r = r.json()
        for i in range(len(r)):
            V = r[i]["properties"]
            G = r[i]["geometry"]["coordinates"]
            wrapper.append(earthquake_usgs.EarthquakeUSGS(V["mag"], G[0], G[1], V["place"], V["title"], V["url"], V["time"]))
    else:
        r = requests.get(url=latest_url, params=str(PARAMS))
        data = r.json()
        D = data["Earthquakes"]
        for i in range(len(D)):
            V = D[i]["properties"]
            G = D[i]["geometry"]["coordinates"]
            wrapper.append(earthquake_usgs.EarthquakeUSGS(V["mag"], G[0], G[1], V["place"], V["title"], V["url"], V["time"]))
    return wrapper


def get_shakespeare_data(endpoint = "", textonly = False):
    """
    @brief Get data of Shakespeare works (plays, poems)

    This function retrieves  and formats the data into a
    a list of Shakespeare objects.

    Valid endpoints: 'poems','plays'
    Valid queryParams: format{simple}

    @throws Exception if the request fails

    Args: 
        endpoint:  can be either "plays" or "poems". If this is
            specified, then only these types of works are retrieved.
        textonly:  if this is set, then only the text is retrieved.

    Returns:
        An array of Shakespeare objects
    """

    wrapper = []
    url = "http://bridgesdata.herokuapp.com/api/shakespeare/"
    PARAMS = {"Accept: application/json"}

    if endpoint == "plays" or endpoint == "poems":
        url += "/" + endpoint
    if textonly:
        url += "?format=simple"

    r = requests.get(url = url, params = str(PARAMS))
    r = r.json()

    D = r["data"]
    for i in range(len(D)):
        V = D[i]
        wrapper.append(shakespeare.Shakespeare(V["title"], V["type"], V["text"]))
    return wrapper


def get_gutenberg_book_data(num = 0):

    """
    @brief Get meta data of the Gutenberg book collection (1000 books)

    This function retrieves,  and formats the data into a list of
    GutenbergBook objects

    @throws Exception if the request fails

    Returns:
        A list of GutenbergBook objects,
    """
    wrapper = []
    url = "http://bridgesdata.herokuapp.com/api/books"
    PARAMS = {"Accept: application/json"}

    if num > 0:
        url += "?limit=" + str(num)

    r = requests.get(url=url, params = str(PARAMS))
    r = r.json()

    D = r["data"]
    for i in range(len(D)):
        V = D[i]

        A = V["author"]
        L = V["languages"]

        lang = []

        for j in range(len(L)):
            lang.append(str(L[j]))

        G = V["genres"]
        genre = []
        for j in range(len(G)):
            genre.append(str(G[j]))

        S = V["subjects"]
        subject = []
        for j in range(len(S)):
            subject.append(str(S[j]))

        M = V["metrics"]
        wrapper.append(gutenberg_book.GutenbergBook(A["name"], A["birth"], A["death"], V["title"],
                                                    str(lang), str(genre), str(subject), M["characters"], M["words"],
                                                    M["sentences"], M["difficultWords"], V["url"], V["downloads"]))

    return wrapper


def get_cancer_incident_data(num = 0):
    """
    @brief Retrieves the CDC cancer incidence dataset.

    Returns a list of records.  See CancerIncidence class for more information

    Args:
        num:  count of records to be retrieved
    Returns:
         cancer incidence data records in a list
    """
    wrapper = []
    url = "https://bridgesdata.herokuapp.com/api/cancer/withlocations"
    PARAMS = {"Accept: application/json"}

    if num > 0:
        url += "?limit="+str(num)

    r = requests.get(url=url, params=str(PARAMS))
    r = r.json()

    D = r["data"]

    # c = CancerIncidence.CancerIncidence()
    for i in range(len(D)):
        c = cancer_incidence.CancerIncidence()
        v = D[i]
        age = v["Age"]
        c.set_age_adjusted_rate(age["Age Adjusted Rate"])
        c.set_age_adjusted_ci_lower(age["Age Adjusted CI Lower"])
        c.set_age_adjusted_ci_upper(age["Age Adjusted CI Upper"])
        c.set_year(v["Year"])

        data = v["Data"]

        c.set_crude_rate(data["Crude Rate"])
        c.set_crude_rate_ci_lower(data["Crude CI Lower"])
        c.set_crude_rate_ci_upper(data["Crude CI Upper"])
        c.set_race(data["Race"])
        c.set_population(data["Population"])
        c.set_event_type(data["Event Type"])
        c.set_affected_area(v["Area"])

        loc = v["loc"]

        c.set_location_x(loc[0])
        c.set_location_y(loc[1])

        wrapper.append(c)
    return wrapper


def get_song(songTitle, artistName = None):
    """
    @brief Get data of a particular song (including lyrics) using the Genius API.

    Data from Genius API(https://docs.genius.com/), given the song title and artist name.
    Valid endpoints:  http://bridgesdata.herokuapp.com/api/songs/find/
    Valid queryParams: song title, artist name

    This function retrieves  and formats the data into a
    Song object. The song if not cached in the local DB is queried
    and added to the DB

    Args:
        songTitle: title of song
        artistName: name of artist

    @throws Exception if the request fails

    Returns:
         a Song object,
    """
    wrapper = []
    url = "http://bridgesdata.herokuapp.com/api/songs/find/"
    PARAMS = {"Accept: application/json"}

    if len(songTitle):
        url = url + urllib.parse.quote(songTitle)

    if artistName is not None:
        if len(artistName):
            url += "?artistName=" + urllib.parse.quote(artistName)

    r = requests.get(url = url, params = str(PARAMS))
    if r.status_code != 200:
        raise ConnectionError("HTTP Request Failed. Error Code: " + r.status_code)
    r = r.json()
    if "artist" in r:
        artist = r["artist"]
    else:
        artist = ""
    if "song" in r:
        songs = r["song"]
    else:
        songs = ""
    if "album" in r:
        album = r["album"]
    else:
        album = ""
    if "lyrics" in r:
        lyrics = r["lyrics"]
    else:
        lyrics = ""
    if "release_date" in r:
        release_date = r["release_date"]
    else:
        release_date = ""

    return song.Song(artist, songs, album, lyrics, release_date)


def get_song_data():
    """

    Get data of the songs (including lyrics) using the Genius API.

    Song data from https://docs.genius.com/.
    Valid endpoints:  https://bridgesdata.herokuapp.com/api/songs/

    This function retrieves  and formats the data into a list of
    Song objects. This version of the API retrieves all the cached
    songs in the local DB.

    @throws Exception if the request fails

    Returns:
         A list of Song objects,
    """
    all_songs = []
    url = "http://bridgesdata.herokuapp.com/api/songs/"
    params = {"Accept: application/json"}

    r = requests.get(url=url, params=str(params))
    r = r.json()

    D = r["data"]

    for i in range(len(D)):
        v = D[i]

        if "artist" in v:
            artist = v["artist"]
        else:
            artist = ""

        if "song" in v:
            song = v["song"]
        else:
            song = ""

        if "album" in v:
            album = v["album"]
        else:
            album = ""

        if "lyrics" in v:
            lyrics = v["lyrics"]
        else:
            lyrics = ""

        if "release_date" in v:
            release_date = v["release_date"]
        else:
            release_date = ""

        all_songs.append(song.Song(artist, song, album, lyrics, release_date))
    return all_songs


def get_color_grid_from_assignment(server: str, user: str, assignment: int, subassignment: int = 0) -> ColorGrid:
    """
    @brief Reconstruct a ColorGrid from an existing ColorGrid on the bridges server

    This method can be useful in early CS courses like CS1 or CS2  to 
    work with an existing data structure holding a dataset, like an image.

    Args:
        server:  server holding the assignment
        user:  user name
        assignment:  assignment number
        subassignment:  sub-assignment number

    Returns:
        ColorGrid structure
    """

    response = get_assignment(server, user, assignment, subassignment)

    try:
        from types import SimpleNamespace as Namespace
    except ImportError:
        # Python 2.x fallback
        from argparse import Namespace

    assignment_object = json.loads(response, object_hook=lambda d: Namespace(**d))
    try:
        if assignment_object.assignment_type != "ColorGrid":
            raise RuntimeError("Malformed ColorGrid JSON: not a ColorGrid")

        data_list = assignment_object.data
        if len(data_list) != 1:
            raise RuntimeError("Malformed JSON: data is malformed")
        data = data_list[0]
        try:
            encoding = data.encoding
        except AttributeError:
            # Handle case for ColorGrid generated before encoding field was added
            encoding = "RAW"

        if encoding != "RLE" and encoding != "RAW":
            raise RuntimeError("Malformed ColorGrid JSON: encoding not supported: " + encoding)

        if len(data.dimensions) != 2:
            raise RuntimeError("Malformed ColorGrid JSON: dimensions are malformed")
        dim_x = int(data.dimensions[0])
        dim_y = int(data.dimensions[1])

        try:
            node_string = str(data.nodes)
        except (TypeError, ValueError):
            raise RuntimeError("Malformed ColorGrid JSON: node is not a String")

        import base64
        decoded_bytes = bytearray(base64.b64decode(node_string))
        decoded = [int(x) for x in decoded_bytes]
    except AttributeError:
        raise RuntimeError("Malformed JSON: Unable to Parse. Does this assignment exist?")

    color_grid = ColorGrid(dim_x, dim_y)

    if encoding == "RAW":
        if len(decoded) != dim_x * dim_y * 4:
            raise RuntimeError("Malformed ColorGrid JSON: nodes is not the size we expect for RAW encoding")

        base = 0
        for x in range(0, dim_x):
            for y in range(0, dim_y):
                color = Color(decoded[base],
                              decoded[base + 1],
                              decoded[base + 2],
                              (decoded[base + 3]/255.0)
                              )
                color_grid.set(x, y, color)
                base = base + 4

    elif encoding == "RLE":
        if len(decoded) % 5:
            raise RuntimeError("Malformed ColorGrid JSON: RLE nodes are not a multiple of 5")
        current_in_decoded = 0
        current_in_cg = 0
        while current_in_decoded != len(decoded):
            repeat = decoded[current_in_decoded]

            color = Color(decoded[current_in_decoded + 1],
                          decoded[current_in_decoded + 2],
                          decoded[current_in_decoded + 3],
                          (decoded[current_in_decoded + 4] / 255.0)
                          )
            current_in_decoded = current_in_decoded + 5

            while repeat >= 0:
                pos_x = int(current_in_cg / dim_y)
                pos_y = int(current_in_cg % dim_y)
                if pos_x >= dim_x or pos_y >= dim_y:
                    raise RuntimeError("Malformed ColorGrid JSON: Too much data in nodes")

                color_grid.set(pos_x, pos_y, color)
                current_in_cg = current_in_cg + 1
                repeat = repeat - 1

    return color_grid


def get_assignment(server: str, user: str, assignment: int, subassignment: int = 0) -> str:
    """
    @brief This method retrieves a stored assignment from the server.

    Args:
        server:  server holding the assignment
        user: the name of the user who uploaded the assignment
        assignment:  assignment number
        subassignment:  sub-assignment number

    Returns:
        data structure representation of the assignment as a JSON string
    """
    subassignment_fixed = "0{}".format(subassignment) if subassignment < 10 else subassignment
    url = "{}/assignmentJSON/{}.{}/{}".format(server, assignment, subassignment_fixed, user)
    params = "Accept: application/json"

    request = requests.get(url, params)
    if request.ok:
        return request.content
    else:
        raise request.raise_for_status()


def osm_server_request(url):
    """
    @brief Fulfills a server request for OpenStreetMap

    For internal use only

    Args:
        url: url for server to process

    """
    request = requests.get(url)
    if not request.ok:
        if request.status_code == 404:
            request = requests.get(url)
            if not request.ok:
                raise request.raise_for_status()
            raise RuntimeError("Location: {} is not supported,\n valid names: {}".format(location, valid_names))
        raise request.raise_for_status()

    server_data = request.content

    return server_data


def get_osm_data(*args) -> OsmData:
    """
    @brief This method retrieves an OpenStreet Map dataset, given a location
    by name (string).

    Args:
        args: location name

    Returns:
        OSM data
    """
    import os

    debug = True
    
    if (len(args) == 2):
        location = args[0]
        level = args[1]
        url = _get_osm_baseurl()+"/loc?location=" + urllib.parse.quote(location) + "&level=" + urllib.parse.quote(level)
        hash_url = _get_osm_baseurl()+"/hash?location=" + urllib.parse.quote(location) + "&level=" + urllib.parse.quote(level)
    elif (len(args) == 5):
        minLat = str(args[0])
        minLon = str(args[1])
        maxLat = str(args[2])
        maxLon = str(args[3])
        level = args[4]
        url = _get_osm_baseurl()+"/coords?minLon=" + minLon + "&minLat=" + minLat + "&maxLon=" + maxLon + "&maxLat=" + maxLat + "&level=" + level
        hash_url = _get_osm_baseurl()+"/hash?minLon=" + minLon + "&minLat=" + minLat + "&maxLon=" + maxLon + "&maxLat=" + maxLat + "&level=" + level
    else:
        raise RuntimeError("Invalid Map Request Inputs")

    lru = lru_cache.lru_cache(30)

    if debug:
        print ("url: "+url)
        print ("hash_url: "+hash_url)

    try:
        from types import SimpleNamespace as Namespace
    except ImportError:
        from argparse import Namespace


    data = None
    not_skip = True
    hash = osm_server_request(hash_url).decode('utf-8')
    if (hash != "false" and lru.inCache(hash)):
        not_skip = False
        data = lru.get(hash)

    if not_skip:
        content = osm_server_request(url)
        try:
            data = json.loads(content.decode('utf-8'), object_hook=lambda d: Namespace(**d))
        except:
            print("Error: Corrupted JSON download...\nAttempting redownload...")
            content = osm_server_request(url)
            try:
                data = json.loads(content.decode('utf-8'), object_hook=lambda d: Namespace(**d))
            except Exception as e:
                print(f"Error: Redownload attempt failed\n{e}")
                sys.exit(0)

        hash = osm_server_request(hash_url).decode('utf-8')
        lru.put(hash, data)


    try:
        if data.nodes is None or data.edges is None or data.meta is None:
            raise RuntimeError("Malformed OSM JSON")

        vertex_map = {}
        vertices = []
        for i, node in enumerate(data.nodes):
            vertex_map[node[0]] = i
            vertices.append(OsmVertex(node[1], node[2]))

        edges = []
        for edge in data.edges:
            id_from = edge[0]
            id_to = edge[1]
            dist = edge[2]
            edges.append(OsmEdge(vertex_map[id_from], vertex_map[id_to], dist))

        ret_osm = OsmData()
        ret_osm.edges = edges
        ret_osm.vertices = vertices
        ret_osm.name = data.meta.name
        return ret_osm

    except AttributeError:
        raise RuntimeError("Malformed JSON: Unable to parse")

def server_request(url):
    """
    @brief Fulfills a server request
    
    For internal use only
    """
    request = requests.get(url)
    if not request.ok:
        if request.status_code == 404:
            request = requests.get(url)
            if not request.ok:
                raise request.raise_for_status()
            raise RuntimeError("404 issue with server request")
        elif request.status_code == 504:
            request = requests.get(url)
            if not request.ok:
                raise request.raise_for_status()
            raise RuntimeError(f"504 issue with server request: {request.raise_for_status}")
        raise request.raise_for_status()

    server_data = request.content

    return server_data

def get_elevation_data(*args):
    """
    @brief This method retrieves an Elevation  Map dataset, given a location
        by name (string).
    Args:
        args(0): a bounding box, aka an array [minLat, minLon, maxLat, maxLon]
        args(1): spatial resolution, aka the distance between two samples (in degrees)

    Returns:
        Elevation data for the bounding box and resolution requested 
        (approximately) [type: ElevationData]
    """
    base_url = get_elevation_url() + "/elevation"
    hash_url = get_elevation_url() + "/hash"

    coords = args[0]
    minLat = str(coords[0])
    minLon = str(coords[1])
    maxLat = str(coords[2])
    maxLon = str(coords[3])

    res = .0166
        
    if len(args) == 2:
        res = args[1]
        
    url = base_url + f"?minLat={minLat}&minLon={minLon}&maxLat={maxLat}&maxLon={maxLon}&resX={res}&resY={res}"
    hash_url = hash_url + f"?minLat={minLat}&minLon={minLon}&maxLat={maxLat}&maxLon={maxLon}&resX={res}&resY={res}"
    #loads cache
    lru = lru_cache.lru_cache(30)

    data = None
    not_skip = True
    hash = False
    hash = server_request(hash_url).decode('utf-8')
    if (hash != "false" and lru.inCache(hash)):
        not_skip = False
        data = lru.get(hash)

    if not_skip:
        data = server_request(url).decode("utf-8")

    hash = server_request(hash_url).decode('utf-8')
    lru.put(hash, data)

    
    #parse and build object
    ret_ele = ElevationData()
    
    file_array = data.splitlines()
    ret_ele.cols = int(file_array[0].split(" ")[-1])
    ret_ele.rows = int(file_array[1].split(" ")[-1])
    ret_ele._xll = float(file_array[2].split(" ")[-1])
    ret_ele._yll = float(file_array[3].split(" ")[-1])
    ret_ele.cellsize = float(file_array[4].split(" ")[-1])

    maxVal = float('-inf')
    minVal = float('inf')
    x = 5
    while (x < len(file_array)):
        arr = file_array[x].replace("\n", "").split(" ")
        arr.pop(0)
        parsedline = []
        for y in arr:
            parsedline.append(int(y))
            if(int(y) > maxVal):
                maxVal = int(y)
            if(int(y) < minVal):
                minVal = int(y)
        ret_ele.data.append(parsedline)
        x += 1
    ret_ele.maxVal = maxVal
    ret_ele.minVal = minVal    

    return ret_ele

def _get_wiki_actor_movie_direct(year_begin, year_end, array_out):
    """
    @brief get actor movie data from wiki data
    Args:
        year_begin: beginning year of data
        year_end: ending year of data.
        array_out: returned data in an array
    """
    # allocate cache - this makes no sense
    lru = lru_cache.lru_cache(30)

    # check cache code name and check cache
    code_name = "wikidata-actormovie-" + str(year_begin) + "-" + str(year_end)
    print(code_name)
    from_cache = False
    if lru.inCache(code_name):  
        results = lru.get(code_name)
        from_cache = True
    else:
        sparql = SPARQLWrapper("https://query.wikidata.org/sparql")
        sparql.setQuery("""
            SELECT ?movie ?movieLabel ?actor ?actorLabel WHERE \
            {\
                ?movie wdt:P31 wd:Q11424.\
                ?movie wdt:P161 ?actor.\
                ?movie wdt:P364 wd:Q1860.\
                ?movie wdt:P577 ?date.\
                FILTER(YEAR(?date) >= """ + str(year_begin) + """ && YEAR(?date) <= """ + str(year_end) + """).\
                SERVICE wikibase:label { bd:serviceParam wikibase:language \"en\". } \
            }
        """)
        sparql.addCustomHttpHeader("User-Agent", 'bridges-python')
        sparql.setReturnFormat(JSON)
        results = sparql.query().convert()
        lru.put(code_name, results)

    for result in results["results"]["bindings"]:
        mak = movie_actor_wiki_data.MovieActorWikiData()
        actor_uri = str(result['actor']['value'])
        movie_uri = str(result['movie']['value'])
        actor_uri = actor_uri.replace("http://www.wikidata.org/entity/","",1)
        movie_uri = movie_uri.replace("http://www.wikidata.org/entity/","",1)
        mak.actor_uri = actor_uri
        mak.movie_uri = movie_uri
        mak.movie_name = str(result['movieLabel']['value'])
        mak.actor_name = str(result['actorLabel']['value'])
        array_out.append(mak)
        # print(result['movie']['value'])

def get_wiki_data_actor_movie(year_begin, year_end):
    """
    @brief This method retrieves an actor-movie data from Wikidata

    Uses a sparql query

    Args:
        year_begin:  beginning year of data request
        year_end:    ending year of data request

    Returns:
        A list of the actor-movie data of type MovieActorWikiData
    """
    ret = []
    for y in range(year_begin, year_end+1):
        _get_wiki_actor_movie_direct(y, y, ret)
    return ret

def get_amenity_data(*args):
    """
    @brief This method retrieves amenity data from Open Street Map datasets.
    
    Args:
        [0]: minimum latitude
        [1]: minimum longitude
        [2]: maximum latitude
        [3]: maximum longitude
        [4]: amenity type
      or
        [0]: location string
        [1]: amenity type

    Valid types are "food", "school", "firestation", "airport", "heli"

    Returns:
        A list of amenities
    """
    
    if(len(args)) == 5:
        url = f"{get_amenity_url()}/amenity?minLat={args[0]}&minLon={args[1]}&maxLat={args[2]}&maxLon={args[3]}&amenity={args[4]}"
        hash_url = f"{get_amenity_url()}/hash?minLat={args[0]}&minLon={args[1]}&maxLat={args[2]}&maxLon={args[3]}&amenity={urllib.parse.quote(args[4])}"
    elif(len(args) == 2):
        url = f"{get_amenity_url()}/amenity?location={urllib.parse.quote(args[0])}&amenity={urllib.parse.quote(args[1])}"
        hash_url = f"{get_amenity_url()}/hash?location={args[0]}&amenity={args[1]}"
    else:
        raise RuntimeError("Invalid Number of Map Request Inputs")

    lru = lru_cache.lru_cache(30)
    data = None
    not_skip = True

    hash = osm_server_request(hash_url).decode('utf-8')
    if (hash != "false" and lru.inCache(hash)):
        not_skip = False
        data = lru.get(hash)
    
    if (not_skip):
        content =  osm_server_request(url)
        try:
            data = json.loads(content.decode('utf-8'))
        except:
            print("Error: Corrupted JSON download...\nAttempting redownload...")
            content = osm_server_request(url)
            try:
                data = json.loads(content.decode('utf-8'))
            except Exception as e:
                print(f"Error: Redownload attempt failed\n{e}")
                raise RuntimeError(e)
                sys.exit(0)

        hash = osm_server_request(hash_url).decode('utf-8')
        lru.put(hash, data)


    ret_data = amenities()


    for i, node in enumerate(data['nodes']):
        temp = amenityData()
        for x, vals in enumerate(node):
            if (x == 0):
                temp.id = vals
            elif (x == 1):
                temp.lat = vals
            elif (x == 2):
                temp.lon = vals
            elif (x == 3):
                temp.name = vals
            else:
                temp.addOther = vals
        ret_data.data = temp
    
    temp_meta = meta(data['meta']['minlat'], data['meta']['minlon'], data['meta']['maxlat'], data['meta']['maxlon'], data['meta']['count'])
    ret_data.meta = temp_meta

    return ret_data

def get_gutenberg_book_metadata(*args):
    """
    @brief function to search for a gutenberg book given a search string and type of metadata to search through
    :param args:
        args[0]: search string, i.e home
        args[1]: metadata type, i.e id, title, lang, date_added, authors, genres, loc_class
    :return: a list of books with containing matched strings from the specified metadata type
    """
    url = get_gutenberg_url() + f"/search?search={args[0]}&type={args[1]}"

    content = server_request(url)
    data = json.loads(content.decode('utf-8'))

    book_list = []
    for node in data['book_list']:
        meta = gutenberg_meta.GutenbergMeta ()

        meta.id = node["id"]
        meta.title = node['title']
        meta.lang = node["lang"]
        meta.date = node["date_added"]
        meta.authors = node["authors"]
        meta.genres = node["genres"]
        meta.loc = node["loc_class"]

        book_list.append(meta)

    return book_list

def get_a_gutenberg_book_metadata(id):
    """
    @brief function to retrieve the metadata of a gutenberg book given its ID
    :param id: ID of the book
    :return: a dictionary containing the metadata of the gutenberg book
    """
    url = get_gutenberg_url() + "/meta?id=" + str(id)

    content = server_request(url)
    data = json.loads(content.decode('utf-8'))

    meta = gutenberg_meta.GutenbergMeta
    for node in data['book_list']:
        
        
        meta.id = node["id"]
        meta.title = node['title']
        meta.lang = node["lang"]
        meta.date = node["date_added"]
        meta.authors = node["authors"]
        meta.genres = node["genres"]
        meta.loc = node["loc_class"]

    return meta

def gutenberg_book_text(id, strip = False):
    """
    @brief function to retrieve the text of a gutenberg book given the ID
    :param id: id of the book
    :param strip: boolean to determine if headers and footers are stripped from the text
    :return: json containing the text of the book
    """
    url = get_gutenberg_url() + "/book?id=" + str(id)
    
    lru = lru_cache.lru_cache(120)
    try:
        if (lru.inCache("gutenberg" + str(id))):
            if source_type == "local" or source_type == "testing":
                print(f"Using cache to request {id}")
            data = lru.get("gutenberg" + str(id))
        else:
            if source_type == "local" or source_type == "testing":
                print(f"Using {source_type} server to request {id}")
            content = server_request(url)
            data = content.decode('utf-8')
            lru.put("gutenberg" + str(id), data)
    except Exception as e:
        print(e)
        raise RuntimeError(e)

    book_data = json.loads(data)

    return book_data

def reddit_data(subreddit, time_request = -9999):
    """
    @brief function to retrieve the metadata of a gutenberg book given its ID
    :param subreddit: the name of the subreddit
    :param time_request: unix timestamp of when requested subreddit was generated
    :return: a list of reddit objects with the data of the posts
    """
    base_url = get_reddit_url()
    url = f"{base_url}/cache?subreddit={subreddit}&time_resquest={time_request}"




    content = server_request(url)
    data = json.loads(content.decode("utf-8"))

    reddit_posts = []

    for n in data:
        post = reddit.reddit
        post.id = data[n]["id"]
        post.title = data[n]["title"]
        post.author = data[n]["author"]
        post.score = int(data[n]["score"])
        post.vote_ratio = int(data[n]["vote_ratio"])
        post.comment_count = int(data[n]["comment_count"])
        post.subreddit = data[n]["subreddit"]
        post.post_time = int(data[n]["post_time"])
        post.url = data[n]["url"]
        post.text = data[n]["text"]

        reddit_posts.append(post)




    return reddit_posts



class DataSource:
    pass



from bridges.graph_benchmark import *
from bridges.line_chart import *
from datetime import datetime
import time as time_


class PageRankBenchmark(GraphBenchmark):
    def __init__(self, p):
        super(PageRankBenchmark, self).__init__()
        self.__plot = p
        self.__plot.x_label = "Number of Edges"
        self.__plot.y_label = "Runtime (in s)"

    def run(self, algoname, paralgo):
        time = []
        vtx_count = []
        edge_count = []

        years = 0
        while years < 120:
            year = 2019 - years
            graph = GraphAdjList()
            vertex_count, edge_count = self._genertate_wiki_data_movie_actor(year, 2019, graph)
            pr = dict()
            start = int(round(time_.time() * 1000))
            paralgo(graph, pr)
            end = int(round(time_.time() * 1000))
            elapsed_time = end - start

            time.append(elapsed_time)
            vtx_count.append(vertex_count)
            edge_count.append(edge_count)

            if elapsed_time > self.time_cap:
                break
            years = 1.2 * years + 1

        self.__plot.set_x_data(algoname, edge_count)
        self.__plot.set_y_data(algoname, time)


from bridges.polyline import *


class Polygon(Polyline):

    def __init__(self, pts = None):
        """
        Constructor. Sets the number of points in the polygon
        Args:
            pts: number of points to be used in constructing the polygon
        """
        if pts is not None:
            super(Polygon, self).__init__(pts = pts)
        else:
            super(Polygon, self).__init__()

    def get_shape_type(self):
        """
        Get the name of this symbol
        """
        return "polygon"




import requests
import traceback

class Connector:
    """
    @brief This is a class for handling calls to the BRIDGES server to transmit JSON to the server and subsequent visualization. It is not intended for external use
    """

    server_url_live = "http://bridges-cs.herokuapp.com"
    server_type = "application"
    server_url_clone = "http://bridges-clone.herokuapp.com"
    server_url_local = "http://127.0.0.1:3000"
    server_url_game = "http://bridges-games.herokuapp.com"

    key = ""
    username =""

    pattern_found = 0
    debug = False

    def __init__(self, key, username, assignment):
        """
        @brief Connector object constructor
        Args:
            key: is the user_key
            username: is students username
            assignment: is the assignment number for the assignment
        """
        self.key = key
        self.username = username
        self.assignment = assignment
        self.server_url = "http://bridges-cs.herokuapp.com"

    def set_server(self, server):
        """
        @brief Set the server based on a keyword for url
        Args:
            server: is one of the three string keywords('live', 'clone', 'local') that is passed to change the server that the bridges visualization will be sent to
        """
        self.set_server_url(server)


    def set_server_url(self, server_url):
        """
        @brief Set the server url (string) that is passed to change the server that the bridges visualization will be sent to
   
        Args: 
            server: is one of the three strings ('live', 'clone', 'local')
        """
        switcher = {
            "live": "http://bridges-cs.herokuapp.com",
            "clone": "http://bridges-clone.herokuapp.com",
            "local": "http://127.0.0.1:3000",
            "games": "http://bridges-games.herokuapp.com"
        }
        self.server_url = switcher.get(server_url, ValueError("Use: live, clone, local to determine url"))


    def get_server_url(self):
        """
        @brief Get the server url (string)
        Returns:
            return server url (string)
        """
        return self.server_url

    def post(self, url, data):
        """
        @brief post the data (JSON) to the server
        Args:
            url:  url of server
            data: JSON of the data structure representation
        """
        try:
            if self.key.isdigit() is not True:
                raise Exception("Key entered is not a valid Key. Please enter a valid Key")
        except Exception as e:
            return traceback.print_tb(e.__traceback__)

        r = requests.post(self.prepare(url), headers={u'content-type': u'application/json'}, data=data.encode('utf-8'))
        if r.status_code != 200:
             print(r.status_code, r.reason)
             print(r.text)
        return r.status_code

    def prepare(self, url):
        """
        @brief prepare the  post string for transmission to the server
        Args:
            url:  url of server
        """
        out = self.server_url
        out += url
        out += "?apikey=" + self.key + "&username=" + self.username
        return out




from enum import IntEnum


class NamedSymbol(IntEnum):
    none = 0
    A = 1
    B = 2
    C = 3
    D = 4
    E = 5
    F = 6
    G = 7
    H = 8
    I = 9
    J = 10
    K = 11
    L = 12
    M = 13
    N = 14
    O = 15
    P = 16
    Q = 17
    R = 18
    S = 19
    T = 20
    U = 21
    V = 22
    W = 23
    X = 24
    Y = 25
    Z = 26
    a = 27
    b = 28
    c = 29
    d = 30
    e = 31
    f = 32
    g = 33
    h = 34
    i = 35
    j = 36
    k = 37
    l = 38
    m = 39
    n = 40
    o = 41
    p = 42
    q = 43
    r = 44
    s = 45
    t = 46
    u = 47
    v = 48
    w = 49
    x = 50
    y = 51
    z = 52
    zero = 53
    one = 54
    two = 55
    three = 56
    four = 57
    five = 58
    six = 59
    seven = 60
    eight = 61
    nine = 62
    empty0 = 63
    circle = 64
    square = 65
    diamond = 66
    triangle_down = 67
    triangle_up = 68
    triangle_left = 69
    triangle_right = 70
    man = 71
    woman = 72
    cat = 73
    pawn = 74
    knight = 75
    bishop = 76
    rook = 77
    queen = 78
    king = 79
    arrow_right = 80
    arrow_left = 81
    arrow_up = 82
    arrow_down = 83
    star = 84
    droplet = 85
    heart = 86
    lightning = 87
    thoughtbubble = 88
    speechbubble = 89
    flower = 90
    square_thatch = 91
    circle_x = 92
    circle_x2 = 93
    circle_segment = 94
    empty1 = 95
    sword = 96
    horn_helmet = 97
    astro_helmet = 98
    target = 99
    empty2 = 100
    waves = 101
    rain = 102
    drink = 103
    paperclip = 104
    elephant = 105
    cowboyhat = 106
    ballcap = 107
    flag = 108
    bomb = 109
    bear = 110
    giraffe = 111
    bug1 = 112
    bug2 = 113
    bug3 = 114
    magnifying_glass = 115
    fishes = 116
    duck = 117
    sloth = 118
    bird = 119
    apple = 120
    carrot = 121
    lemon = 122
    pepper = 123
    onion = 124
    potion1 = 125
    potion2 = 126
    bow_and_arrow = 127
    campfire = 128
    donut = 129
    monitor = 130


#!/usr/bin/env python
from decimal import Decimal
from bridges.element import *
from bridges.color import *
from decimal import Decimal

class ElementVisualizer():

    """
    @brief This class is used to store the visualization elements on the for the bridges
    Visualization, including the color, shape, opacity, and size of the node.<p>
    
    <b>Supported named colors: CSS colors. 
    Check https://drafts.csswg.org/css-color-3/#svg-color</b>: <p>
    
    <b> Color by RGBA Specification :</b>  Range: 0-255 for each component <p>
    
    <b> Thickness: </b> Range : 0.0-50.0
    
    <b> Opacity: </b> Range (0.0-1.0) </p>
    
    <b> Supported Shapes: </b> "circle", "square", "diamond", "cross", "triangle", "star", "wye". 
    
    Objects of this class are stored as properties of all Element subclasses.
    Generally, you will manipulating the ElementVisualizer returned from the
    Element getVisualizer() method, and then call the setVisualizer() method on
    the Element after changes have been made.
    
    @author Matthew Mcquaigue
    
    @date 2018, 6/24/19
    """

    def __init__(self, color="green", shape ="circle", size=10.0, opacity=1.0):
        """
        ElementVisualizer constuctor
        kwargs:
            (str) color: color for this element
            (str) shape: the shape of the element
            (float) size: the element size
            (float) opacity: the element opacity
        Returns:
            None
        """
        self.prop = dict()
        self._color = Color(70, 130, 180, 1.0)
        if color != "green":
            self.color = color
        self._shape = shape
        self._size = size
        self._opacity = opacity
        self._locationX = Decimal("Infinity")
        self._locationY = Decimal("Infinity")
        self.prop['color'] = ["70", "130", "180", "1.0"]
        self.prop["opacity"] = "1.0"
        self.prop["size"] = "10.0"
        self.prop["shape"] = "circle"
        self.prop["key"] = ""
        self.prop["locationX"] = Decimal("Infinity")
        self.prop["locationY"] = Decimal("Infinity")


    @property
    def size(self) -> float:
        """
        Getter for the elements size
        Returns:
            float: the size
        """
        return self._size

    @size.setter
    def size(self, size) -> None:
        """
        Setter for the size of the element
        Args:
            (float) size: the elements desired size
        Returns:
            None
        """
        self.prop["size"] = str(size)
        self._size = size

    @property
    def color(self) -> Color:
        """
        Getter for the color of the element in the bridges visualization
        Returns:
            Color: Color object representing the color of the element
        """
        return self._color

    @color.setter
    def color(self, *args, **kwargs) -> None:
        """
        Setter for the color of the element in the bridges visualization
        Args:
            (optional) list: requires either 3 ints 0-255 for RGB and an optional float 0.0-1.0 for alpha EX: color = [0, 255, 0, 1.0]
            (optional) str: string representing the element color. from web colors: https://developer.mozilla.org/en-US/docs/Web/CSS/color_value
        Returns:
            None
        Raises:
            ValueError: if the color name provided is not available
        """
        self._color = Color(*args, **kwargs)

    @property
    def shape(self) -> str:
        """
        Getter for the shape of the element
        Returns:
            str: reperesenting the type of shape
        """
        return self._shape

    @shape.setter
    def shape(self, a_shape):
        """
        Setter for the shape of the element
        Args:
            (str) a_shape: the name of shape for element
        """
        a_shape = a_shape.lower()
        self.prop["shape"] = a_shape
        self._shape = a_shape

    @property
    def opacity(self) -> float:
        """
        Getter for the opacity of the element
        Returns:
            float: representing the opacity
        """
        return self.color.alpha

    @opacity.setter
    def opacity(self, opacity) -> None:
        """
        Setter for the opacity of the element
        Args:
            (float) opacity: the opacity to be applied
        Returns:
            None
        """
        self.prop["opacity"] = Decimal(opacity)
        self.color.alpha = opacity
        self._opacity = opacity

    def set_location(self, x, y):
        """
        Setter for the location of the element
        Args:
            (int) x: x location
            (int) y: y location
        Returns:
            None
        """
        self.prop['locationX'] = x
        self.prop['locationY'] = y
        self._locationX = x
        self._locationY = y

    @property
    def location_x(self):
        """
        Getter for the X location of element
        Returns:
            int: as the x location
        """
        return self._locationX

    @property
    def location_y(self):
        """
        Getter for the y location of the element
        Returns:
            int: as the y position
        """
        return self._locationY



#!/usr/bin/env python
from typing import Any, Union

from bridges.sl_element import *
from bridges.edge import *
import traceback


class GraphAdjList:
    """
    @brief The GraphAdjList class can be used to represent adjacency list based graphs in BRIDGES
    
    The GraphAdjList class can be used to represent adjacency list based  graphs
    in BRIDGES
    
    Convenience methods are provided to add vertices and edges to 
    the graph as well as
    retrieve the adjacency list of a vertex, given its id.
    Convenience method addVertex() is provided to add vertices to
    the graph, and addEdge() is provided to add edges.  Edges are
    retrieved by using the dual hashmap, given the vertex ids of the
    edge. Vertices can be styled directly from the vertex element
    returned by getVertex(), and edges are styled from a LinkVisualizer
    one can access through getLinkVisualizer(). Here is a simple example:
    
    \code{java}
    graph = GraphAdjList()
    graph.add_vertex("a");
    graph.add_vertex("b");
    graph.add_edge("a", "b");
    graph.get_vertex("a").set_shape("square");
    graph.get_link_visualizer("a", "b").set_color("yellow");
    \endcode
    
    Adjacency lists are singly linked lists using the BRIDGES
    SLelement. Iterators are provided for easy traversal of the
    adjacency lists. For instance,
    
    \code{java}
    graph = GraphAdjList()
    for (e : graph.out_going_edge_set_of("a"))
         print("a -> "+e.tov());
    \endcode
        
    Graphs can have their nodes and links affected by visual attributes. Nodes
    can have color, size, opacity and shape and  detailed in the ElementVisualizer
    class. Edges support attributes such as color, thickness and opacity and are
    detailed in the LinkVisualizer class.  Element and link attributes are set
    using the get_visualizer() and get_link_visualizer() methods.  For instance,
    
    \code{java}
    GraphAdjList graph = something();
    graph.add_vertex("baskin");
    graph.add_vertex("robins");
    graph.add_edge("baskin","robins");
    graph.get_visualizer().set_color("cyan");
    graph.get_visualizer().set_shape("square");
    graph.get_link_visualizer("baskin", "robins").set_color("green");
    graph.get_link_visualizer("baskin", "robins").set_opacity("0.5f");
    \endcode
    
    
    @author Matthew Mcquaigue, Kalpathi Subramanian
    @date 2018,  7/23/19, 1/5/21
    
    \sa graph adjacency list tutorial, https://bridgesuncc.github.io/tutorials/Graph.html
    
    There are two visualization engines available for graph. The small
    graph visualization supports all attributes of vertices and edges
    but is prohibitively slow on large graphs. The large graph
    visualization only supports locations (actually they are mandatory)
    and colors, all other attributes are ignored.
    
    BRIDGES picks the rendering engine automatically. But it can be
    forced to pick one used forceLargeVizualization() and
    forceSmallVizualization
    """

    LargeGraphVertSize = 1000
    force_large_viz = False
    force_small_viz = False

    def __init__(self) -> None:
        """
        Constructor for a graph adj list
        Returns:
            None
        """
        self._vertices = dict()
        self.adj_list = dict()

    def get_data_structure_type(self) -> str:
        """
        Getter for the data structure type
        Returns:
            str: representing the type
        """
        if (GraphAdjList.force_large_viz == True or (GraphAdjList.force_small_viz == False and
                                                     self.LargeGraphVertSize < len(self.vertices) and
                                                     self.are_all_vertices_located())):
            return "largegraph"
        return "GraphAdjacencyList"

    def add_vertex(self, id: str,
        label: str = None,
        data: Any = None,
        color: Union[str, Color] = "blue",
        opacity: float = 1.0,
        original: Element = None
       ) -> Element:
       """
       Adds a new vertex to the graph. Initializes the adjacency list; user is responsible for checking if the vertex already exists.  This method will replace the value for this key.  Note: it is the user's responsibility to  check for duplicate vertices
     
       Args:
            id: the vertex id
            label: vertex label
            data:  the vertex data
            color:  color of vertex
            opacity:  opacity of vertex
            original:  an existing Element instance to copy, this will 
                 nullify other arguments
       Returns:
            newly created Vertex
       """
       self.vertices[id] = Element(val=data, color=color, opacity=opacity, original=original)
       self.vertices.get(id).label = str(label) if label is not None else str(id)
       self.adj_list[id] = None

       return self.vertices[id]

    def add_edge(self,
                 src: Union[str, Element],
                 dest: Union[str, Element],
                 label: str = None,
                 data: any = None,
                 color: Union[str, Color] = "blue",
                 thickness: float = 1.0,
                 opacity: float = 1.0
                 ) -> Edge:
        """
        Adds a new edge to the graph. Adds edge to that vertex's adjacency list; user is responsible for checking if the vertex already exists. This version assumes a default edge weight of 1.
        Args:
            src: source vertex of edge
            dest: destination  vertex of edge
            label: label of edge,
            color: color of edge,
            thickness: thickness of edge,
            opacity: opacity of edge,
            data: data the edge will hold
        Returns:
            newly created Edge
        Raises:
            ValueError: if the src and dest vertices do not exist
        """
        source_id = self._resolve_id(src)
        dest_id = self._resolve_id(dest)
        if source_id is None or dest_id is None:
            raise ValueError("Vertex " + src + " or " + dest +
                             " does not exist! Add the vertex before creating the edge.")
        try:
            if self.vertices.get(source_id) is None or self.vertices.get(dest_id) is None:
                raise ValueError("Vertex " + src + " or " + dest +
                                 " does not exist! Add the vertex before creating the edge.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        self.adj_list[source_id] = SLelement(e=Edge(source_id, dest_id, data, label=label, color=color,
                                                    thickness=thickness), next=self.adj_list.get(src))

        return self.adj_list[source_id].value

    def set_vertex_data(self, src, vertex_data) -> None:
        """
        Set for the data for a given vertex.
        Args:
            src: the source vertex
            vertex_data: The data for the vertex
        Returns:
            None
        Raises:
             ValueError: if the source vertex doesnt exist
        """
        try:
            if self.vertices[src] is None:
                raise ValueError("Vertex " + src + " does not exist!")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        self.vertices[src].value = vertex_data

    def get_vertex_data(self, src):
        """
        Get the data stored as part of this vertex.
        Args:
            src: the source vertex
        Returns:
            data stored at this vertex
        """
        try:
            if self.vertices[src] is None:
                raise ValueError("Vertex " + src + " does not exist!")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return self.vertices[src].value

    def set_edge_data(self, src, dest, edge_data):
        """
        Set edge specific data for the edge.
        Args:
            src: source vertex of edge
            dest: destination vertex of edge
            edge_data: edge specific data
        """
        try:
            if self.vertices[src] is None or self.vertices[dest] is None:
                raise ValueError("Vertex " + src + " or " + dest + " does not exist!")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        sle = self.adj_list[src]
        while sle is not None:
            edge_dest = sle.value.tov
            if edge_dest == dest:
                if edge_data is not None:
                    sle.value.set_edge_data(edge_data)
                    return
                else:
                    return sle.value.edge_data
            sle = sle.next
        if sle is None:
            raise ValueError("VEdge from " + src + " to " + dest + "does not exist!")

    @property
    def vertices(self) -> dict:
        """
        Getter for the graph nodes
        Returns:
            dict
        """
        return self._vertices

    def get_vertex(self, key):
        """
        Getter for a specific vertex in the dictionary of vertices.
        Args:
            key: The associated key for the vertex
        Returns:
            vertex: Element or None
        """
        return self.vertices.get(key)

    def get_adjacency_list(self, vertex=None):
        """
        Gets the adjacency list
        Args:
            vertex: input  vertex 
        Returns:
            list : adjacency list of this vertex
        """
        if vertex is not None:
            return self.adj_list.get(vertex)
        else:
            return self.adj_list

    def key_set(self):
        """
        Gets the keys for all the graph vertices
		Returns:
			the keys for the vertices in the graph; used with iterators
		"""
        return self.vertices.keys()

    def value_set(self):
        return self.vertices.values()

    def out_going_edge_set_of(self, v):
        """
        Get the outgoing set of edges from a given vertex
        Args:
           v: the given vertex 
        Returns:
           outgoing edge set (used in iterartors)
        """
        return SLelement.list_helper(start=self.get_adjacency_list(v))

    def get_edge_data(self, src, dest):
        """
        Get edge specific data from this edge.

        Args:
            src: source vertex of edge
            dest: destination vertex of edge
        Returns:
            data stored on this edge
        """
        sle = self.adj_list[src]
        while sle is not None:
            ed = sle.value
            if ed.destination == dest:
                return ed.edge_data
            sle = sle.next

    def get_edge(self, src: Union[str, Element], dest: Union[str, Element]):
        """
        Get the edge given its vertices.

        Args:
            src: source vertex of edge
            dest: destination vertex of edge
        Returns:
            edge corresponding to the source and dest vertices
        """
        source_id = self._resolve_id(src)
        dest_id = self._resolve_id(dest)
        if source_id is None or dest_id is None:
            raise ValueError("Vertex " + src + " or " + dest +
                             " does not exist! Add the vertex before creating the edge.")

        sle = self.adj_list[source_id]
        while sle is not None:
            ed = sle.value
            if ed.destination == dest_id:
                return ed
            sle = sle.next

    def _resolve_id(self, key_or_element: Union[str, Element]) -> Union[str, None]:
        if type(key_or_element) is str:
            return key_or_element
        elif type(key_or_element) is Element:
            return [key for key, value in self.vertices.items() if value == key_or_element].pop()
        else:
            return key_or_element

    def are_all_vertices_located(self):
        """
        Check if all vertices in the graph have a valid location
        Returns:
            a boolean
        """
        for element in self.vertices.items():
            el = element[1]
            elvis = el.visualizer
            if(elvis.location_x == float('inf') or elvis.location_y == float('inf')):
                return False
        return True

    def force_large_visualization(self, f):
        """
        Force the rendering engine to use large graph visualization. This forces the rendering to a more bandwidth efficient at the cost of having less features. The large graph visualization only renders vertices that have specified locations. The only usable attribute for vertices and edges are colors.

        Args:
          f: boolean set to true to force the visualization engine to use large graphs visualization. Setting to false does not prevent large visualization to be used, just does not force it.
        Returns:
          None

        """
        if f:
            GraphAdjList.force_large_viz = True
            GraphAdjList.force_small_viz = False
        else:
            GraphAdjList.force_large_viz = False

    def force_small_visualization(self, f):
        """
        Force the rendering engine to use small graph visualization
        
        The small visualization uses more bandwidth, have more
        features, and support a force directed layout for vertices
        which do not have a specified location.

        Args:
           f: boolean set to true to force the visualization engine to use small graphs visualization. Setting to false does not prevent small visualization to be used, just does not force it.
        Returns:
            None

        """
        if f:
            GraphAdjList.force_small_viz = True
            GraphAdjList.force_large_viz = False
        else:
            GraphAdjList.force_small_viz = False

    def get_link_visualizer(self, src, dest):
        """
        This is a convenience method to simplify access to the link visualizer; the method assumes the vertex names point to existing vertices, else an exception is thrown
       Args:
            src - source vertex of edge
            dest - destination vertex of edge
        """
        #  get the source and destination vertex elements
        #  and check to see if they exist
        return self.get_edge(src, dest)

    def get_visualizer(self, vertex):
        """
        This is a convenience method to simplify access to the element visualizer.
	
        The method assumes the vertex name points to an existing vertex, else an exception is thrown
       Args:
            vertex - The vertex for which visualizer is wanted
       """
        v = self.vertices.get(vertex)
        try:
            if v is None:
                raise ValueError("Vertex " + vertex + " does not exist! First add the vertices to the graph.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return v.visualizer

    def get_data_structure_representation(self) -> dict:
        """
        Get the data structure representation
        Returns:
              dict, a JSON representation of the structure 
        """
        node_map = dict()  # map to reorder the nodes for building JSON
        nodes = []  # get the list nodes
        nodes_JSON = []  # array for list of nodes in json
        links_JSON = []  # array for building the links JSON - traverse the adj. lists

        # redirect for large graphs
        if (GraphAdjList.force_large_viz == True or (GraphAdjList.force_small_viz == False and
                                                     self.LargeGraphVertSize < len(self.vertices) and
                                                     self.are_all_vertices_located())):
            return self.get_data_structure_large_graph()

        # get the objects and add them to the array
        for elements in self.vertices.items():
            nodes.append(elements[1])

        # append all nodes representation to list of nodes
        for i in range(0, len(nodes)):
            node_map[nodes[i]] = i
            nodes_JSON.append(nodes[i].get_element_representation())

        # get all nodes in adj_list
        for a_list in self.adj_list.items():
            links_list = a_list[1]
            src_vert = self.vertices.get(a_list[0])
            #  get the source vertex index for the JSON (int)
            while links_list is not None:
                src_indx = node_map.get(src_vert)
                #  get the destination vertex index for the JSON (int)
                edge = links_list.value
                dest_vert = self.vertices.get(edge.tov)
                dest_indx = node_map.get(dest_vert)
                #  get link representation
                links_JSON.append((links_list.get_link_representation(
                                   edge._lvis,
                                   str(src_indx),
                                   str(dest_indx))))
                links_list = links_list.next
        json_str = {
            "nodes": nodes_JSON,
            "links": links_JSON
        }
        return json_str

    def get_data_structure_large_graph(self) -> dict:
        """
        Gets the JSON of the large graph representation
        Returns:
            the data structure representation 
        """
        nodes = []
        node_map = dict()

        for element in self.vertices.items():
            nodes.append(element[1])

        nodes_json = []
        for k in range(0, len(nodes)):
            nodes_json.append([])
            node_map[nodes[k]] = k
            elvis = nodes[k].visualizer
            loc_str = []
            if elvis.location_x != float('inf') and elvis.location_y != float('inf'):
                loc_str.append(elvis.location_x)
                loc_str.append(elvis.location_y)
                nodes_json[k].append(loc_str)
            color = elvis.color
            nodes_json[k].append([color.red, color.green, color.blue, color.alpha])

        links_json = []
        for a_list in self.adj_list.items():
            list = a_list[1]
            src_vert = self.vertices.get(a_list[0])
            while list is not None:
                link_json=[]
                src_indx = node_map.get(src_vert)
                edge = list.value
                dest_vert = self.vertices.get(edge.tov)
                dest_indx = node_map.get(dest_vert)
                color = edge.color
                link_json.append(src_indx)
                link_json.append(dest_indx)
                link_json.append([color.red, color.green, color.blue, color.alpha])
                links_json.append(link_json)
                list = list.next

        graph_alist_json = {
            "nodes": nodes_json,
            "links": links_json
        }

        return graph_alist_json



#!/usr/bin/env python
from bridges.element import *
import copy


class SLelement(Element):
    """
    @brief This class can be used to instantiate Singly Linked Elements.
    
    This class extends Element and takes a generic parameter <E>
    representing application specific data. This element forms the basic
    building block for singly linked lists. Singly linked elements have a field
    pointing to the next element along the list.
    
    Elements contain a visualizer (ElementVisualizer) object for setting visual
    attributes (color, shape, opacity, size), necessary for displaying them in a
    web browser.
    
    Elements also have a LinkVisualizer object, that is used when they are linked to
    another element, appropriate for setting link attributes, for instance, between
    the current element and its next element.
    
    @author Matthew Mcquaigue
    @date  2018, 6/24/19, 2021
    
    \sa Singly Linked List tutorial : https://bridgesuncc.github.io/tutorials/SinglyLinkedList.html
    """

    def __init__(self, **kwargs) -> None:
        """
        Conctructor for SLelement object
        Kwargs:
            (Generic) e: the generic object that this SLelement will hold
            (str) label: the label of the SLelement that shows up on the bridges visualization
            (Element) next: the element that should be assigned to the next pointer
        Returns:
            None
        """
        super(SLelement, self).__init__(val=kwargs.get('e'), label=kwargs.get('label'))
        self._next = kwargs.get('next')


    def get_data_structure_type(self) -> str:
        """
        Getter for the data structure type
        Returns:
            str: representing the data structure type
        """
        return "SinglyLinkedList"


    @property
    def next(self):
        """
        Getter for element following this element
        Returns:
            SLelement
        """
        return self._next

    @next.setter
    def next(self, n) -> None:
        """
        Setter for the element following this element
        Args:
            n: the element to be assigned to next
        Returns:
            None
        """
        self._next = n
        if n is not None:
            self.set_link_visualizer(n)

    @property
    def value(self):
        """
        Getter for the SLelement value to hold
        Returns:
            Element
        """
        return super(SLelement, self).value

    @value.setter
    def value(self, val):
        """
        Setter for the value that this SLelement will hold
        Args:
            val: the value that this SLelment will hold
        Returns:
            None
        """
        super(SLelement, self).value = val

    def list_helper(start):
        """
        helper method for graph adjacency list
        """
        if start is None:
            return
        node = start
        while node.next is not None:
            yield node.value
            node = node.next
        yield node.value

    def get_data_structure_representation(self) -> dict:
        """
        Getter for this data structure representation
        Returns:
            dict: representing the JSON before dumping
        """
        node_map = dict()# map to reorder the nodes for building JSON
        nodes = []# get the list nodes
        self.get_list_elements(nodes)
        nodes_JSON = []# generate the JSON of the list nodes
        k = 0
        while k < len(nodes):
            node_map[nodes[k]] = k
            nodes_JSON.append(nodes[k].get_element_representation())
            k += 1

        links_JSON = []
        k = 0
        while k < len(nodes):
            par = nodes[k]
            chld = par.next
            if chld is not None:
                #  add the link
                links_JSON.append(self.get_link_representation(par.get_link_visualizer(chld),
                                                               str(node_map.get(par)), str(node_map.get(chld))))
            k += 1
        #add the link
        json_dict = {
            "nodes": nodes_JSON,
            "links": links_JSON
        }
        return json_dict



    def get_list_elements(self, nodes):
        """
        Get the elements of the list (for  internal use only)
        Args:
            nodes: a vector of the nodes in the list
        Returns:
            elements of list
        """
        el = self
        #try to handle all lists in subclasses, except multilists
        nodes.clear()
        while el is not None:
            nodes.append(el)
            el = el.next
            #  handle circular lists
            if el == self:
                break

    def iterator(self):
        """
        used for range loops
        """
        return SLelementIterator(self)


class SLelementIterator():

    def __init__(self, current):
        self.current = current

    def has_next(self):
        return self.current is not None

    def next(self):
        if not self.has_next():
            raise StopIteration

        ret = self.current.value
        self.current = self.current.next
        return ret



class Song:
    """
    @brief  A Song object, used along with the Songs data source.

    This is a convenience class provided for  users who wish to use this
    data source as part of their application. It provides an API that makes
    it easy to access the attributes of this data set.

    This object is generally not created by the user, to see how its created check 
    out bridges::data_src_dependent::data_source::get_song()
    
    For an example, check out https://bridgesuncc.github.io/tutorials/Data_Song_Lyrics.html

    @author Matthew Mcquaigue, Kalpathi Subramanian
    @date   2018, 12/29/20
    """

    def __init__(self, artist: str = "", song: str = "", album: str = "", lyrics: str = "", release_date: str = ""):
        """
        @brief Constructor
        Args:
            artist: song artist
            song: song title
            album: song album
            lyrics: lyrics of song
            release_date: release date of song
        """
        self._artist = artist
        self._song = song
        self._lyrics = lyrics
        self._album = album
        self._release_date = release_date

    @property
    def artist(self):
        """
        @brief return artist of song
        Returns:
            artist name of song
        """
        return self._artist

    @artist.setter
    def artist(self, a):
        """
        @brief Set artist of song
        Args:
            a: artist name to set
        """
        self._artist = a

    @property
    def song_title(self):
        """
        @brief return title of song
        Returns:
            song title
        """
        return self._song

    @song_title.setter
    def song_title(self, s):
        """
        @brief Set the song title
        Args:
            s: artist name to set
        """
        self._song = s

    @property
    def album_title(self):
        """
        @brief return album title
        Returns:
            album title of song
        """
        return self._album

    @album_title.setter
    def album_title(self, a):
        """
        @brief Set title of song
        Args:
            a: album title to set
        """
        self._album = a

    @property
    def lyrics(self):
        """
        @brief return lyrics of song
        Returns:
            lyrics of song
        """
        return self._lyrics

    @lyrics.setter
    def lyrics(self, l):
        """
        @brief Set artist of song
        Args:
            l:  lyrics data to set
        """
        self._lyrics = l

    @property
    def release_date(self):
        """
        @brief return release date of song
        Returns:
            release date of song
        """
        return self._release_date

    @release_date.setter
    def release_date(self, r):
        """
        @brief Set release date of song
        Args:
            r: release date to set
        """
        self._release_date = r







class Primitives:

    def __init__(self, type: str, position: list = None):
        self._type = type
        self._position = position
        self._object_json = {
            'type': self._type

        }

    def push_representation(self, scene_json):
        scene_json['meshes'].append(self._object_json)


from bridges.dl_element import *

class CircDLelement(DLelement):
    """
    @brief This class can be used to instantiate Circular Doubly Linked List Elements.
    
    Structurally they are the same as doubly linked elements
    except that each node constructed with the next and the previous pointers
    points to itself.
    
    User's implementation of the circularly linked list needs to ensure that
    the last node's next pointer points to the first node and the first node's
    previous pointer points to the last node, as the visualization generation
    is dependent on this.
    
    Elements have labels (string) that are displayed on the visualization.
    Elements take an generic object E as a user defined parameter, which can be
    any native type or object.
    
    Elements contain a visualizer (ElementVisualizer) object for setting visual
    attributes (color, shape, opacity, size), necessary for displaying them in a web
    browser.
    
    Elements also have a LinkVisualizer object that is used when they are
    linked to another element, appropriate for setting link attributes, between
    the element and its previous or next nodes.
    
    @author Kalpathi Subramanian, Mathhew McQuaigue
    
    @date   7/17/16, 1/16/17 7/23/19, 2021
    
    Circular doubly linked list tutorial: https://bridgesuncc.github.io/tutorials/CircularDoublyLinkedList.html
    """


    def __init__(self, **kwargs) -> None:
        """
        Constructor for an Circularly Doubly Linked Element
        Args:
            label: THe label for this CircDLelement
            e: the generic element object that this CircDLelement will hold
            next: the next DLelement that should be assigned to the next pointer
            prev: THe previous DLelement that should be assigned to the next pointer
        Returns:
            None
        """
        if 'e' in kwargs:
            if 'next' in kwargs and 'prev' in kwargs:
                if 'label' in kwargs:
                    super(CircDLelement, self).__init__(e=kwargs['e'], label=kwargs['label'],
                                                        next=kwargs['next'], prev=kwargs['prev'])
                else:
                    super(CircDLelement, self).__init__(e=kwargs['e'], next=kwargs['next'],
                                                        prev=kwargs['prev'])
            elif 'label' in kwargs:
                super(CircDLelement, self).__init__(e=kwargs['e'], label = kwargs['label'])
                self.next = self.prev = self
            else:
                super(CircDLelement, self).__init__(e=kwargs['e'])
                self.next = self.prev = self
        else:
            super(CircDLelement, self).__init__()
            self.next = self.prev = self

    def get_data_structure_type(self) -> str:
        """
        Gets the data structure type
        Returns:
            None
        """
        return "CircularDoublyLinkedList"

    @property
    def next(self):
        """
        Getter for the next element of this CircDLelement
        Returns:
            DLelement: the following element
        """
        return super(CircDLelement, self).next

    @next.setter
    def next(self, next) -> None:
        """
        Setter for the next element for this CircDLelement
        Args:
            (DLelement) next: the next element to be set
        Returns:
            None
        """
        DLelement.next.fset(self, next)

    @property
    def prev(self):
        """
        Getter for the prev element of this CircDLelement
        Returns:
            DLelement: the prev element in the list
        """
        return super(CircDLelement, self).prev

    @prev.setter
    def prev(self, prev) -> None:
        """
        Setter for the prev element of this CircDLelement
        Args:
            (DLelement) prev: The prev element to be set
        Returns:
            None
        """
        DLelement.prev.fset(self, prev)

    def iterator(self):
        """
        list iterator to be used with range loops 
        """
        return CircDLelementIterator(self)

    def reverse_iterator(self):
        """
        list reverse iterator to be used with range loops 
        """
        return CircDLelementReverseIterator(self)

class CircDLelementIterator():
    def __init__(self, current):
        self.current = current
        self.first = current
        self.at_start = True

    def has_next(self):
        if ((self.current == self.first) and not self.at_start):
            return False

        return True

    def next(self):
        ret = self.current.value
        self.current = self.current.next
        self.at_start = False
        return ret

class CircDLelementReverseIterator():

    def __init__(self, current):
        self.current = current
        self.first = current
        self.at_start = True

    def has_next(self):

        if ((self.current == self.first) and not self.at_start):
            return False

        return True

    def next(self):
        ret = self.current.value
        self.current = self.current.prev
        self.at_start = False
        return ret




from bridges.color import *
from bridges.grid import *
import base64

class ColorGrid(Grid):
    """
    @brief This is a class in BRIDGES for representing an (n x n) grid.
    
    A ColorGrid is essentially an image. One can construct an image of
    a particular size using the ColorGrid() constructor to be either
    blank or filled with a particular Color depending on which
    constructor is called.
    
    \code{.py}
    grid = new ColorGrid(rows, columns)
    grid.set(2, 3, Color("lightsalmon")
    \endcode
    
    You can get a ColorGrid from an existing Bridges ColorGrid assignment using
    bridges.get_color_grid_from_assignment(bridges.get_username(), bridges.get_assignment_id(), 0)
    
    @author David Burlinson, Matthew McQuaigue
    
    @date 2018, 7/24/19, 2021
    
    Color grid tutorial at https://bridgesuncc.github.io/tutorials/Grid.html
    """


    baseColor = Color(r=0, g=0, b=0, a=1.0)

    def get_data_structure_type(self) -> str:
        """
        Get the data structure type
        Returns:
            str : data structure type
        """
        return "ColorGrid"

    def __init__(self, rows: int = 10, cols: int = 10, color: Color = baseColor) -> None:
        """
        Color Grid constructor
        Args:
            rows: number of rows in the grid
            cols: number of columns in the grid
            color: base color of the each grid pixel
        Returns:
            None
        """
        super(ColorGrid, self).__init__(rows=rows, cols=cols)
        self.base_color = color
        self.grid_size = [rows, cols]
        self.initialize_grid()

    def initialize_grid(self) -> None:
        """
        initialize the grid anf populate with base colors
        Returns:
            None
        """
        for i in range(self.grid_size[0]):
            for j in range(self.grid_size[1]):
                self.set(i, j, self.base_color)

    def set(self, row: int, col: int, color: Color) -> None:
        """
        Set the (row, col) element in the color grid
        Args:
            row - which row to access
            col - which col to access
            color - background color for the cell at row,col
        Returns:
            None
        """
        super(ColorGrid, self).set(row, col, color)

    def get_rle(self) -> bytearray:
        """
        Get the run length encoding of color grid
        Returns:
            bytearray
        """
        img_bytes = bytearray()
        count = 0
        total_count = 0
        pos = 0
        last = self.grid[0][0]

        while pos < self.grid_size[0] * self.grid_size[1]:
            posY = pos / self.grid_size[1]
            posX = pos % self.grid_size[1]
            current = self.grid[int(posY)][int(posX)]

            if count == 0:
                count = 1
                last = current
            else:
                if last == current:
                    count += 1
                else:
                    total_count += count
                    img_bytes.append(count-1)
                    last = last.get_byte_representation()

                    for k in range(len(last)):
                        img_bytes.append(last[k])

                    count = 1
                    last = current
            if count == 256:
                total_count += count
                img_bytes.append(count-1)
                last = last.get_byte_representation()
                for k in range(len(last)):
                    img_bytes.append(last[k])
                count = 0
            pos += 1
        total_count += count
        img_bytes.append(count-1)
        last = last.get_byte_representation()
        for k in range(len(last)):
            img_bytes.append(last[k])

        if total_count != self.grid_size[0] * self.grid_size[1]:
            print("Something broke in getRLE construction")

        return img_bytes

    def get_raw(self) -> bytearray:
        """
        Get raw encoding of color grid
        Returns:
            bytearray: representing the colors of grid cells
        """
        img_bytes = bytearray()
        for i in range(self.grid_size[0]):
            if self.grid[i] is not None:
                for j in range(self.grid_size[1]):
                    if self.grid[i][j] is not None:
                        color = self.grid[i][j]
                        color = color.get_byte_representation()
                        for k in range(len(color)):
                            img_bytes.append(color[k])
        return img_bytes

    def get_data_structure_representation(self) -> dict:
        """
        Get the JSON representation of the color grid
        Returns:
            str: representing the json  of the color grid
        """
        byte_buff = self.get_rle()
        encoding = "RLE"

        if len(byte_buff) > self.grid_size[0] * self.grid_size[1] * 4:
            encoding = "RAW"
            byte_buff = self.get_raw()
            print("RAW ran")
        else:
            print("RLE ran")

        json_dict = {
            "encoding": encoding,
            "nodes": [base64.b64encode(bytes(byte_buff)).decode()],
            "dimensions": [self.grid_size[0], self.grid_size[1]]
        }

        return json_dict



import math
from bridges.graph_adj_list import *

class OsmEdge:
    """
    @brief  Class that hold Open Street Map edges
   
    Class that holds Open Street Map edges from https://openstreetmap.org

    This object is generally not created by the user, to see how its 
    created check out bridges::data_src_dependent::data_source::get_osm_data()

    @sa For an example, check out https://bridgesuncc.github.io/tutorials/Data_OSM.html

    @author Erik Saule, Matthew Mcquaigue, Jay Strahler, Kalpathi Subramanian
    @date 2019, 12/28/20
    """
    @property
    def source(self) -> int:
        """
        @brief Get source vertex (int)

        Returns:
            source vertex of edge
        """ 
        return self._source

    
    @source.setter
    def source(self, source: int):
        """
        @brief Set source vertex 
        Args:
            source: source vertex to set
        """
        try:
            value = int(source)
        except ValueError:
            raise ValueError("Source must be an int")

        self._source = value

    @source.deleter
    def source(self):
        """
        @brief delete source vertex
        """
        del self._source

    @property
    def destination(self) -> int:
        """
        @brief Get destination vertex

        Returns: 
            destination vertex of edge
        """
        return self._destination

    @destination.setter
    def destination(self, destination: int):
        """
        @brief Set destination vertex 
        Args:
             destination:  destination vertex to set (int)
        """ 
        try:
            value = int(destination)
        except ValueError:
            raise ValueError("Destination must be an int")

        self._destination = value

    @destination.deleter
    def destination(self):
        """
        @brief delete destination vertex
        """
        del self._destination

    @property
    def distance(self) -> float:
        """
        @brief Get distance between two vertices of the edge (float)
        Returns:
            edge length
        """
        return self._distance

    @distance.setter
    def distance(self, distance: float):
        """
        @brief Set distance between source and destination
        Args:
            distance:  distance to set (float)
        """
        try:
            value = float(distance)
        except ValueError:
            raise ValueError("Distance must be a float")

        self._distance = value

    @distance.deleter
    def distance(self):
        """
        @brief Delete distance between edge vertices
        """
        del self._distance


    def __init__(self, source: int, destination: int, distance: float):
        """
        @brief Constructor
        Args:
            source: source vertex of edge
            destination: destination vertex of edge
            distance: distance between the source and destination vertices
        """
        self._source = 0
        self._destination = 0
        self._distance = 0.0
        self.source = source
        self.destination = destination
        self.distance = distance


class OsmVertex:
    """
    @brief  Class that hold Open Street Map vertices
    
    Class that holds Open Street Map vertices from https://openstreetmap.org

    This object is generally not created by the user, to see how its created check 
    out bridges::data_src_dependent::data_source::get_osm_data()

    @sa For an example, check out https://bridgesuncc.github.io/tutorials/Data_OSM.html
    
    @author Jay Strahler, Matthew Mcquaigue, Erik Saule, Kalpathi Subramanian
    @date 2/14/19, 12/29/20
    """

    @property
    def latitude(self) -> float:
        """
        @brief Get latitude of vertex position

        Returns:
            latitude of vertex
        """
        return self._latitude

    @latitude.setter
    def latitude(self, latitude: float):
        """
        @brief Set latitude of vertex position
        Args:
            latitude: vertex latitude to set (float)
        """ 
        try:
            value = float(latitude)
        except ValueError:
            raise ValueError("latitude must be a float")

        self._latitude = value
        self._to_cartesian_coord()

    @latitude.deleter
    def latitude(self):
        """
        @brief Delete latitude of vertex position
        """
        del self._latitude

    @property
    def longitude(self) -> float:
        """
        @brief Get longitude of vertex position

        Returns:
            longitude of vertex
        """
        return self._longitude

    @longitude.setter
    def longitude(self, longitude: float):
        """
        @brief Set longitude of vertex position
        Args:
            longitude: vertex latitude to set (float)
        """ 
        try:
            value = float(longitude)
        except ValueError:
            raise ValueError("longitude must be a float")

        self._longitude = value
        self._to_cartesian_coord()

    ##
    #    @brief Delete longitude of vertex position
    #
    @longitude.deleter
    def longitude(self):
        """
        @brief Delete longitude of vertex position
        """
        del self._longitude

    ## 
    #  @brief convert lat/long to Cartesian coordinates
    #
    def _to_cartesian_coord(self):
        earth_radius = 6378
        lat_rad = self.latitude * math.pi/180
        longit_rad = self.longitude * math.pi/180
        self.cartesian_coord[0] = earth_radius * math.cos(lat_rad) * math.cos(longit_rad)
        self.cartesian_coord[1] = earth_radius * math.cos(lat_rad) * math.sin(longit_rad)

    def __init__(self, latitude: float, longitude: float):
        """OSM vertex, represents vertex using open street map data
        :param latitude: float, latitude of vertex
        :param longitude: float, longitude of vertex
        """
        self._latitude = 0
        self._longitude = 0
        self.cartesian_coord = [0, 0]
        self.latitude = latitude
        self.longitude = longitude
        self._to_cartesian_coord()



class OsmData:
    """
    @brief  Class that hold Open Street Map Data
    
    Class that holds Open Street Map data, from https://openstreetmap.org

    This object is generally not created by the user, to see how its created check 
    out bridges::data_src_dependent::data_source::get_osm_data()

    @sa For an example, check out https://bridgesuncc.github.io/tutorials/Data_OSM.html
    
    @author Jay Strahler, Matthew Mcquaigue, Erik Saule, Kalpathi Subramanian
    @date 2/16/19, 12/28/20, 1/6/21
    """
    from typing import List
    VertexList = List[OsmVertex]
    EdgeList = List[OsmEdge]

    @property
    def vertices(self) -> VertexList:
        """
        @brief get vertices of the dataset (list)
        Returns:
            list of vertices
        """
        return self._vertices

    @vertices.setter
    def vertices(self, vertices: VertexList):
        """
        @brief Set vertices of the dataset (VertexList)
        Args:
             vertices: VertexList  to set
        """
        if not isinstance(vertices, list):
            raise ValueError("vertices must be a list of OsmVertex objects")

        lat_range = [math.inf, -math.inf]
        lon_range = [math.inf, -math.inf]
        cart_range_x = [math.inf, -math.inf]
        cart_range_y = [math.inf, -math.inf]
        # find the ranges of lat/lon and cartesian coordinates from new vertices
        for vertex in vertices:
            if not isinstance(vertex, OsmVertex):
                raise ValueError("vertices must be a list of OsmVertex objects")

            lat = vertex.latitude
            lon = vertex.longitude
            cart_x = vertex.cartesian_coord[0]
            cart_y = vertex.cartesian_coord[1]

            lat_range[0] = lat if lat_range[0] > lat else lat_range[0]
            lat_range[1] = lat if lat_range[1] < lat else lat_range[1]

            lon_range[0] = lon if lon_range[0] > lon else lon_range[0]
            lon_range[1] = lon if lon_range[1] < lon else lon_range[1]

            cart_range_x[0] = cart_x if cart_range_x[0] > cart_x else cart_range_x[0]
            cart_range_x[1] = cart_x if cart_range_x[1] < cart_x else cart_range_x[1]

            cart_range_y[0] = cart_y if cart_range_y[0] > cart_y else cart_range_y[0]
            cart_range_y[1] = cart_y if cart_range_y[1] < cart_y else cart_range_y[1]

        self.latitude_range = lat_range
        self.longitude_range = lon_range
        self.cartesian_range_x = cart_range_x
        self.cartesian_range_y = cart_range_y
        self._vertices = vertices

    @vertices.deleter
    def vertices(self):
        """
        Delete  vertices of the dataset
        """
        del self._vertices

    @property
    def edges(self) -> EdgeList:
        """
        @brief Get edges  of the dataset (EdgeList)
        Returns:
           edges of the OSM data
        """
        return self._edges

    ##
    #    @brief Set edges of the dataset
    #    @return edges (EdgeList)
    @edges.setter
    def edges(self, edges: EdgeList):
        """
        @brief Set edges  of the dataset
        Args:
           edges: edges to set
        """
        if not isinstance(edges, list):
            raise ValueError("edges must be a list of OsmVertex objects")
        for edge in edges:
            if not isinstance(edge, OsmEdge):
                raise ValueError("edges must be a list of OsmEdge objects")
        self._edges = edges

    @edges.deleter
    def edges(self):
        """
        @brief delete the edges in the dataset
        """
        del self._edges

    def get_graph(self) -> GraphAdjList:
        """
        @brief Construct a graph out of the vertex and edge data of the OSM object. 

        The graph will associate the length
        of the edge to the graph edge. No data is bound to the vertices

        The vertices of the graph will be located at
        the location where given in the data set
        converted to cartesian coordinates
 
        Returns:
            adjacency list based graph
        """
        ret_graph = GraphAdjList()
        for k, vertex in enumerate(self.vertices):
            ret_graph.add_vertex(k, data=vertex)
            ret_graph.get_vertex(k).visualizer.set_location(vertex.cartesian_coord[0], vertex.cartesian_coord[1])
            ret_graph.get_vertex(k).visualizer.color = ("green")

        for k, edge in enumerate(self.edges):
            ret_graph.add_edge(edge.source, edge.destination, data=edge.distance)

        return ret_graph

    def __init__(self):
        """
        @brief constructor
        """
        self._vertices = []
        self._edges = []
        self.latitude_range = []
        self.longitude_range = []
        self.cartesian_range_x = []
        self.cartesian_range_y = []
        self.name = None



from bridges.gamebase import *
from bridges.input_helper import *
import time


class NonBlockingGame(GameBase):
    """
	brief This class provides the features necessary to implement  simple non blocking games.

	The games that can be created out of NonBlockingGame are
	based on a simple board grid of at most 1024 cells (e.g.,
	32x32, or any combinations less than 1024 cells). Each
	cell has a background color, and a colored symbol.

	This class is used by having another class derive
	from it and implement the two functions: initialize()
	and GameLoop(). initialize() is called exactly
	once, on the first frame of the game. It is used to
	make first time initializations of the game state
	(such as setting the board in its initial position,
	for instance in chess). However, GameLoop() is called
	at every frame of the game. The game starts when the
	function start() is called on the object you
	created.

	This game does not do anything, but it is the
	minimal code that will run a game. Note that the
	constructor of my_game passes 3 parameters to the
	constructor of NonBlockingGame(). These three
	parameters are the classic parameters that the
	constructor of bridges::Bridges takes (e.g.,
	assignmentID, username, apikey).

	To change the board, two functions are
	used. setBGColor() change the background color of a
	particular cell. It takes three parameters, the
	first two identify the cell of the board to change,
	and the last one is a color from a color palette
	provided by NamedColor. drawSymbol() takes four
	parameters, the first two identify the cell of the
	board to change, the third is a symbol from a
	symbol palette provided by NamedSymbol, the fourth
	is the color that symbol shold be drawn in and
	provided by NamedColor.

	For instance, a very simple initialize() could look like:
	code{.py}
	def initialize():
	    set_bg_color(0, 0, NamedColor.lightsalmon);
	    draw_symbol(0, 0, NamedSymbol.sword, NamedColor.blue);
	endcode

    Note that the size of the board is set by default
	at 10x10 and that drawing on a cell that does not
	exist will lead to an error. One can specify a
	gameboard of a different size by passing additional
	parameters to the NonBlockingGame
	constructor. However, the game can not be more than
	1024 cells in total, so a 15x15 board is possible,
	a 32x32 board is the largest square board possible,
	and a rectangular 64x16 rectangular board is also
	possible. But a 100x20 board would be 2000 cells
	and is not possible. For instance a board of 16
	rows and 64 columns can be created defining the
	my_game constructor as:

	code{.py}
		my_game = NonBlockingGame (1, "myuserid",  "myapikey", 16, 64)
	endcode

	The bridges game engine will call the GameLoop()
	function at each frame of the game. You can write
	this function to modify the state of the game board
	using setBGColor() and drawSymbol(). For instance,
	the following GameLoop() will color the board
	randomly one cell at a time.

	code{.py}
	def game_oop():
		set_BG_Color(rand()%10, rand()%10, NamedColor.lightsalmon);
	endcode

	The gameLoop can also probe the state of some keys
	of the keyboard using functions key_up(), key_left(),
	key_down(), key_right(), key_w(), key_a(), key_s(),
	key_d(), key_space(), and key_q(). These functions
	return a boolean that indicate whether the key is
	pressed at that frame or not. For instance, the
	following code will only color the board randomly
	when the up arrow is pressed.

	code{.py}
	def gameLoop():
		if key_up()
		    set_bg_color(rand()%10, rand()%10, NamedColor.lightsalmon);
	endcode

	@author Erik Saule
	@date 72119

    NonBlockingGame tutorial at: https://bridgesuncc.github.io/tutorials/NonBlockingGame.html
    """

    def __init__(self, assid, login, apikey, rows, cols):
        super(NonBlockingGame, self).__init__(assid, login, apikey, rows, cols)
        if cols*rows > 1024:
            print("ERROR: Number of cells in a non-blocking game grid cannot exceed 32x32 or 1024.")
            exit(1)
				
        self.time_of_last_frame = int(round(time.time() * 1000))
        self.ih = InputHelper()
        self.register_keypress(self.ih)
        self._fps = 30

    @property
    def fps(self):
        """
        Get the frame rate at which the game running.
        Returns:
            The target framerate. The game could be somewhat slower depending on how computationally expensive the gameloop is and on the speed of the network
        """
        return self._fps

    @fps.setter
    def fps(self, frames):
        """
        Set the frame rate at which the game running.
        Args:
            frames: frame rate to be set.
        """
        self._fps = frames

    def sleep_timer(self, timems=None):
        if timems is None:
            timems = 1000
        try:
            time.sleep(timems/1000)
        except InterruptedError:
            quit()

    def control_framerate(self):
        hz = 1./self._fps

        current_time = int(round(time.time() * 1000))
        theoretical_next_frame = self.time_of_last_frame + int(hz*1000)
        wait_time = theoretical_next_frame-current_time

        if wait_time > 0:
            self.sleep_timer(wait_time)

        self.time_of_last_frame = int(round(time.time() * 1000))

    def start(self):
        """
        Call this function from main to start game
        """

        frame = 0

        framelimit = os.getenv("FORCE_BRIDGES_FRAMELIMIT", None)
        if (framelimit):
            framelimit = int(framelimit)
            print("limiting number of frame to " + str(framelimit))

        try:
            self.sleep_timer()

            self.render()

            self.initialize()

            self.game_started = True

            while self.game_started:
                self.game_loop()
                self.render()
                self.control_framerate()
                frame = frame + 1
                if (framelimit and frame > framelimit):
                    print ("frame limit of "+ str(framelimit) +" frames reached")
                    self.quit()
        except:
            self.close()
            raise
        
        self.close()

    def key_left(self):
        """
        Is left currently pressed?
        Returns:
            true if left is currently pressed
        """
        return self.ih.left()

    def key_right(self):
        """
        Is right currently pressed?
        Returns:
            true if right is currently pressed
        """
        return self.ih.right()

    def key_up(self):
        """
        Is up currently pressed?
        Returns:
            true if up is currently pressed
        """
        return self.ih.up()

    def key_down(self):
        """
        Is down currently pressed?
        Returns:
            true if down is currently pressed
        """
        return self.ih.down()

    def key_q(self):
        """
        Is q currently pressed?
        Returns:
            true if q is currently pressed
        """
        return self.ih.q()

    def key_space(self):
        """
        Is space currently pressed?
        Returns:
            true if space is currently pressed
        """
        return self.ih.space()

    def key_w(self):
        """
        Is w currently pressed?
        Returns:
            true if w is currently pressed
        """
        return self.ih.w()

    def key_a(self):
        """
        Is a currently pressed?
        Returns:
            true if a is currently pressed
        """
        return self.ih.a()

    def key_s(self):
        """
        Is s currently pressed?
        Returns:
            true if s is currently pressed
        """
        return self.ih.s()

    def key_d(self):
        """
        Is right currently pressed?
        Returns:
            true d right is currently pressed
        """
        return self.ih.d()



from bridges.array import *

class Array1D(Array):
    """
    @brief This is a class can be used to create 1 dimensional arrays of type Element<E>.
    
    @author 	Kalpathi Subramanian, Matthew McQuaigue
    
    @date  	7/18/19
    
    This class can be used to create 1D arrays of type Element<E>  where E
    is a generic object representing application specific data.
    
    Array1D has iterator semantic to enable range for loops. For instance,
    
    \code{java}
    Array1D<Integer> arr = something();
    for (Integer i : arr)
        System.out.println(i);
    \endcode
    
    Example Tutorial at: https://bridgesuncc.github.io/tutorials/Array.html (1D, 2D, and 3D Array)<br>
    """


    def __init__(self, sz=None):
        """
        @brief Create a 1D array object
        @param sz number of elements in the array
        """
        super(Array1D, self).__init__()
        if sz is not None:
            self.size = sz
            dim = [sz, 1, 1]
            self.set_size(1, dim)
        else:
            self.size = 0



from bridges.named_color import *
from bridges.named_symbol import *

class GameCell:
    """
    This class is used to represent cells in GameGrids in BRIDGES.
    Each cell has a foreground color, background color, and symbol.
    @sa  Refer to the Game tutorial at 
    
    https://bridgesuncc.github.io/tutorials/NonBlockingGame.html
    
    @author David Burlinson, Matthew McQuaigue
    @date 9/06/18
    """

    def __init__(self, **kwargs):
        """
        Constructor for the cell of a gamegrid
        Kwargs:
            0: background color from enum named_colors
            1: foreground color from enum named_colors
            2: the symbol in the cell from named_symbols
        Returns:
              None
        """
        if 'bg' in kwargs:
            self._bg = kwargs['bg']
        else:
            self._bg = NamedColor.black
        if 'fg' in kwargs:
            self._fg = kwargs['fg']
        else:
            self._fg = NamedColor.cyan
        if 'symbol' in kwargs:
            self._symbol = kwargs['symbol']
        else:
            self._symbol = NamedSymbol.none

    @property
    def bg(self):
        """
        Getter for the background color
        Returns:
            color: background color
        """
        return self._bg

    @bg.setter
    def bg(self, color):
        """
        Setter for the background color
        Args:
            color: color to be set
        Returns:
            None
        """
        if type(color) == NamedColor:
            self._bg = color
        else:
            self._bg = NamedColor[color]

    @property
    def fg(self):
        """
        Getter for the foreground color
        Returns:
            color: the foreground color
        """
        return self._fg

    @fg.setter
    def fg(self, color):
        """
        Setter for the color of the foreground
        Args:
            color: the color to be applied
        Returns:
            None
        """
        if type(color) == NamedColor:
            self._fg = color
        else:
            self._fg = NamedColor[color]

    @property
    def symbol(self):
        """
        Getter for the symbol in the cell
        Returns:
            symbol
        """
        return self._symbol

    @symbol.setter
    def symbol(self, s):
        """
        Setter for the symbol in the cell
        Args:
            s: the symbol to be applied
        Returns:
            None
        Raises:
            ValueError: if the int value for the symbol is < 0 or > 255
        """
        if type(s) == NamedSymbol:
            self._symbol = s
        else:
            if s < 0 or s > 255:
                raise ValueError("Symbol " + s + " is invalid; symbols must be specified from the range (0, 255)")
            self._symbol = NamedSymbol[s]

    def get_bg_byte(self):
        """
        Gets the background color as a byte.
        Returns:
            background color as byte (index of value in NamedColor)
        """
        return bytes([self._bg.value])

    def get_fg_byte(self):
        """
        Gets the foreground color as a byte.
        Returns:
            foreground color as byte (index of value in NamedColor)
        """
        return bytes([self._fg.value])

    def get_symbol_byte(self):
        """
        Gets the symbol as a byte.
        Returns:
             symbol as byte
        """
        return bytes([self._symbol.value])





from bridges.sl_element import *

class CircSLelement(SLelement):
    """
    @brief This class can be used to instantiate Singly Linked Circular List Elements.
    
    Structurally they are the same as singly linked elements
    except that each node constructed with the next point pointing to itself;
    User's implementation of the circularly linked list needs to ensure that
    the last node points to first node of the list, as the visualization generation is
    dependent on this.
    
    Elements have labels (string) that are displayed on the visualization.
    Elements take an generic object as a user defined parameter, E, which
    can be any native type or object.
    
    Elements contains a visualizer (ElementVisualizer) object for setting visual
    attributes (color, shape, opacity, size), necessary for displaying them in a web
    browser.
    
    Elements also have a LinkVisualizer object that is used when they are
    linked to another element, appropriate for setting link attributes, between
    an element and its next element.
    
    @author Kalpathi Subramanian, Matthew McQuaigue
    
    @date 6/22/16, 1/7/17, 5/17/17, 7/23/19, 12/29/20
    
    Circular singly linked list tutorial: https://bridgesuncc.github.io/tutorials/CircularSinglyLinkedList.html
    """

    def __init__(self, **kwargs) -> None:
        """
        The constructor for a Circular Singly Linked Element
        Args:
            e: the generic object that this CircSLelement will hold
            label: The label of this CircSLelement
            next: The CircSLelement that should be assigned to the next pointer
        Returns:
            None
        """
        if 'e' in kwargs:
            if 'label' in kwargs:
                if 'next' in kwargs:
                    super(CircSLelement, self).__init__(e=kwargs['e'], label=kwargs['label'],
                                                        next = kwargs['next'])
                else:
                    super(CircSLelement, self).__init__(e=kwargs['e'], label=kwargs['label'])
                    self.next = self
            else:
                super(CircSLelement, self).__init__(e=kwargs['e'])
                self.next = self
        else:
            super(CircSLelement, self).__init__()
            self.next = self

    def get_data_structure_type(self) -> str:
        """
        Gets the data structure type
        Returns:
            str: representing the data structure type
        """
        return "CircularSinglyLinkedList"

    @property
    def next(self):
        """
        Getter for the next element of this CircSLelement
        Returns:
            CircSLelement:  the element that follows this element
        """
        return super(CircSLelement, self).next

    @next.setter
    def next(self, n) -> None:
        """
        Setter for the next element in Circular list
        Args:
            n: the next element to be set
        Returns:
            None:
        """
        SLelement.next.fset(self, n)

    def iterator(self):
        """
        list iterator  for use in range loops
        """
        return CircSlelementIterator(self)

class CircSlelementIterator():

    def __init__(self, current):
        self. current = current
        self. first = current
        self.at_start = True

    def has_next(self):
        if((self.current == self.first) and not self.at_start):
            return False

        return True

    def next(self):
        ret = self.current.value
        self.current = self.current.next
        self.at_start = False
        return ret



from bridges.graph_adj_list import *
from bridges.data_src_dependent.data_source import *
import sys

class GraphBenchmark:

    def __init__(self):
        self._time_cap = sys.float_info.max

    def _genertate_wiki_data_movie_actor(self, year_min, year_max, movie_graph):
        v = get_wiki_data_actor_movie(year_min, year_max)
        edge_count = 0
        vertex_count = 0

        for ma in v:
            movie_vertex = movie_graph.get_vertex(ma.movie_name)
            if movie_vertex is None:
                movie_graph.add_vertex(ma.movie_uri, ma.movie_name)
                movie_vertex = movie_graph.get_vertex(ma.movie_uri)
                vertex_count += 1

            actor_vertex = movie_graph.get_vertex(ma.actor_uri)
            if actor_vertex is None:
                movie_graph.add_vertex(ma.actor_uri, ma.actor_name)
                actor_vertex = movie_graph.get_vertex(ma.actor_uri)
                vertex_count += 1

            movie_graph.add_edge(ma.movie_uri, ma.actor_uri)
            movie_graph.add_edge(ma.actor_uri, ma.movie_uri)
            edge_count += 2

        return vertex_count, edge_count

    def _highest_degree_vertex(self, gr):
        max_degree = -1
        ret = ""

        for k in gr.key_set():
            degree = 0
            for e in gr.out_going_edge_set_of(k):
                degree += 1
            if degree > max_degree:
                max_degree = degree
                ret = k

        return ret

    def _count_vertices(self, gr):
        return len(gr.vertices)

    def _count_edges(self, gr):
        edges = 0
        for k in gr.key_set():
            for e in gr.out_going_edge_set_of(k):
                edges+=1
        return edges

    @property
    def time_cap(self):
        return self._time_cap

    @time_cap.setter
    def time_cap(self, cap_in_s):
        self._time_cap = cap_in_s











class Scene:

    def __init__(self):
        self._scene_list = []

    def get_data_structure_type(self):
        return 'Scene'

    def add(self, scene_object):
        self._scene_list.append(scene_object)

    def get_data_structure_representation(self):
        scene_json = {
            'meshes': [],
            'camera': None,
            'lights': []
        }
        for i in range(len(self._scene_list)):
            self._scene_list[i].push_representation(scene_json)

        return scene_json


import json
import math
from bridges.symbol import *

class Label(Symbol):
    """
    @brief This class provides support for text labels to be associated with shapes. 
    
    Labels have width, height, font size.
    
    \sa Shape collection tutorial, https://bridgesuncc.github.io/tutorials/Symbol_Collection.html
    """
    def __init__(self, label = None):
        super(Label, self).__init__()
        if label is not None:
            super(Label, self).label = label
        self._width = 100
        self._height = 50
        self._font_size = 12.
        self._rotation_angle = 0.

    @property
    def width(self) -> int:
        """
        Getter for the width of label
        Returns:
            int: the width
        """
        return self._width

    @width.setter
    def width(self, w: int) -> None:
        """
        Setter for the width of label
        Args:
            (int) w: the width
        Returns:
            None
        """
        self._width = w

    @property
    def height(self) -> int:
        """
        Getter for the height of label
        Returns:
            int: the height
        """
        return self.height

    @height.setter
    def height(self, h: int) -> None:
        """
        Setter for the height of label
        Args:
            (int) h: the height
        Returns:
            None
        """
        self.height = h

    @property
    def font_size(self) -> int:
        """
        Getter for the font size
        Returns:
            int: representing the font size
        """
        return self._font_size

    @font_size.setter
    def font_size(self, size: int) -> None:
        """
        Setter for the font size
        Args:
            size: the size of the font
        Returns:
            None
        Raises:
            ValueError: if the size of the font is <= 0 or > 200
        """
        if size <= 0 or size > 200:
            raise ValueError("Please use font size between 0 and 200")
        else:
            self._font_size = size

    @property
    def rotation_angle(self) -> None:
        """
        Returns the rotation angle of the label

        Returns:
            rotation angle (in degrees)
        """
        return self._rotation_angle

    @rotation_angle.setter
    def rotation_angle(self, angle: float) -> None:
        """
        Set the orientation of the label text
        Only, horizontal, vertical and 45 deg. text is supported

        Args:
            angle: rotation angle (in degrees)
        Returns:
            None
        """
        self._rotation_angle = angle

    def get_bounding_box(self):
        """
        Get the bounding box of the label. Only single line, upper, lower
        case alphabetical characters and spaces supported
        Returns:
            list: representing the label bounding box
        """
        label_text = super(Label, self).label
        length = 0.
        upper_case_exists = False

        for ch in label_text:
            if ch.islower():
                if (ch == 'm' or ch == 'w'):
                    length += 0.6
                elif (ch == 'i' or ch == 'l' or ch == 'j'):
                    length += 0.4
                else: 
                    length += 0.5
            elif ch.isupper():
                upper_case_exists = True
                if (ch == 'M' or ch == 'W'):
                    length += 0.72
                elif  (ch == 'I'): 
                    length += 0.52
                else:
                    length += 0.62
            else:  
                #support spaces only, no special chars
                length += 0.55

        length *= self._font_size;

        width = length
        height = self._font_size
        if upper_case_exists: 
            height += 0.3*self._font_size
        else:
            height += 0.1*self._font_size

        # account for text orientation to compute bounding box 
        bbox_width = length
        bbox_height = height
        angle = self._rotation_angle
        loc_x = self.get_location()[0]
        loc_y = self.get_location()[1] 
        bbox = [0., 0., 0., 0.]
        pt = [0., 0.]
        bbox[0] = bbox[1] = float("inf")
        bbox[2] = bbox[3] = float("-inf")

         # rotate  the four corners of the bounding box
         # only need to rotate the 3 points of the box
        for k in range(4):
             if k == 0:       # lower left
                 pt[0] = pt[1] = 0.
             elif k == 1:     # upper left
                 pt[0] = 0. 
                 pt[1] = bbox_height;
             elif k == 2:     # lower right
                 pt[0] = bbox_width 
                 pt[1] = 0.;
             elif k == 3:     # upper right
                 pt[0] = bbox_width;
                 pt[1] = bbox_height; 

             self.rotate_point (pt, angle);
             # update bounding box
             if pt[0] < bbox[0]: 
                 bbox[0] = pt[0]
             if pt[1] < bbox[1]:
                 bbox[1] = pt[1]
             if pt[0] > bbox[2]:
                 bbox[2] = pt[0]
             if pt[1] > bbox[3]: 
                 bbox[3] = pt[1]

        #translate center of box to center of label
        tx = loc_x - (bbox[0] + (bbox[2]-bbox[0])/2.);
        ty = loc_y - (bbox[1] + (bbox[3]-bbox[1])/2.);
        bbox[0] += tx; 
        bbox[1] += ty; 
        bbox[2] += tx;
        bbox[3] += ty;

        return bbox

    def get_dimensions(self):
        """
        Get the dimensions of the label
        Returns:
            list: representing the label dimensions
        """
        length = 0.09 * self.font_size * len(super(Label, self).label)
        x = self.get_location()[0]
        y = self.get_location()[1]

        return [x - length/2, x + length/2, y, y]

    def get_json_representation(self) -> dict:
        """
        Getter for the json representation
        Returns:
            dict: JSON represnted as a dict before the dump
        """
        ds_json = super(Label,self).get_json_representation()
        ds_json["name"] = super(Label, self).label
        ds_json["shape"] = "text"
        ds_json["angle"] = self._rotation_angle
        ds_json["font-size"] = self.font_size

        return ds_json



import socketio
import json

class SocketConnection:
    """
    This class is used internally as part of the BRIDGES Game API to set up 
    connections to the BRIDGES Game server. Not to be used directly by BRIDGES 
    users directly
    """

    _sio = socketio.Client()
    _listeners = []

    def __init__(self, b):
        self.bridges = b

    @staticmethod
    def add_listener(to_add):
        print("Subscribing to keypress events..", to_add)
        SocketConnection._listeners.append(to_add)

    def setup_connection(self, user, assignment):
        try:
            url = "https://bridges-games.herokuapp.com"
            SocketConnection._sio.connect(url, transports=["websocket"])

            student_cred = {
                'user': user,
                'apikey': self.bridges.get_key(),
                'assignment': self.bridges.get_assignment_id()
            }
            student_cred = json.dumps(student_cred)

            print("passing student credentials to server..")
            SocketConnection._sio.emit('credentials', student_cred)

        except ConnectionError as e:
            print(e)

#    @_sio.on('announcement')
#    def announcement(*args):
#        print("announcement", *args)

    @_sio.on('keydown')
    def keydown(*args):
        #print(*args)
        for i in range(0, len(SocketConnection._listeners)):
            SocketConnection._listeners[i].key_press(*args)

    @_sio.on('keyup')
    def keyup(*args):
        for i in range(0, len(SocketConnection._listeners)):
            SocketConnection._listeners[i].key_press(*args)

    @_sio.on('disconnect')
    def disconnect(*args):
        print(*args, "disconnected")

    def send_data(self, dataframe):
        if SocketConnection._sio is not None:
            data = json.dumps(dataframe)
            SocketConnection._sio.emit('gamegrid:recv', data)

    def close(self):
        SocketConnection._sio.disconnect()
        SocketConnection._sio.wait()











class ElevationData:
    """
    @brief  Class that holds elevation data

    The data is stored as a 2d array .data of .cols columns and .rows rows.
    The data at [0][0] are at location xll, yll and the spatial resolution 
    is cellsize.

    This object is generally not created by the user, to see how its created check 
    out bridges::data_src_dependent::data_source::get_elevation_data()

    @sa For an example, check out https://bridgesuncc.github.io/tutorials/Data_Elevation.html

    @author Jay Strahler, Kalpathi Subramanian
    
    @date 3/28/20, 12/29/20, 1/6/20
    """

    
    @property
    def cols(self):
        """ 
        @brief Get data width
        Returns:
            return the width of the data
        """ 
        return self.ncols

    @cols.setter
    def cols(self, ncols):
        """
        @brief Set width of elevation data grid
        Args:
             ncols:  elevation data width to set
        """
        self.ncols = ncols

    @property
    def rows(self):
        """
        Get data height
        Returns:
            return the height of the data
        """
        return self.nrows

    @rows.setter
    def rows(self, nrows):
        """
        Set height of elevation data grid
        Args: 
           nrows:  dlevation data height to set
        """
        self.nrows = nrows 

    @property
    def data(self):
        """ 
        @brief Get elevation data
        Returns:
            return the elevation data
        """ 
        return self._data

    @data.setter
    def data(self, eledata):
        """
        @brief Set elevation data
        Args: 
          eledata:  elevation data to set
        """
        self._data = eledata

    @property
    def xll(self):
        """ 
        @brief Get xcoord of  data origin
        Returns:
            return the elevation data lower left (x coord)
        """ 
        return self._xll

    @xll.setter
    def xll(self, xcoord):
        """
        @brief Set xcoord of elevation data origin
        Args: 
          xcoord:  elevation data xcoord of origin to set
        """
        self._xll = xcoord

    @property
    def yll(self):
        """ 
        @brief Get ycoord of  data origin
        Returns:
            return the elevation data lower left (y coord)
        """ 
        return self._yll

    @yll.setter
    def yll(self, ycoord):
        """
        @brief Set ycoord of elevation data origin
        Args: 
          ycoord:  elevation data ycoord of origin to set
        """
        self._yll = ycoord

    @property
    def cellsize(self):
        """ 
        @brief Get cell size (resolution)  of  elevation data
        Returns:
            return the elevation data cell size
        """ 
        return self._cellsize

    @cellsize.setter
    def cellsize(self, value):
        """
        @brief Set cell size (resolution)  of elevation data
        Args: 
          value:  cell size to set
        """
        self._cellsize = value
    
    @property
    def maxVal(self):
        """ 
        @brief Get max elevation value in the data
        Returns:
            return the max value in data
        """ 
        return self._maxVal

    @maxVal.setter
    def maxVal(self, value):
        """
        @brief Set max elevation value in the data
        Args: 
          value:  max value to set
        """
        self._maxVal = value


    @property
    def minVal(self):
        """ 
        @brief Get min elevation value in the data
        Returns:
            return the min value in data
        """ 
        return self._minVal

    @minVal.setter
    def minVal(self, value):
        """
        @brief Set min elevation value in the data
        Args: 
          value:  min value to set
        """
        self._minVal = value

        

    def __init__(self):
        """
        @brief Constructor - initialize elevation object
        """
        self.ncols = 0
        self.nrows = 0
        self._data = []
        self._xll = 0
        self._yll = 0
        self._cellsize = 0
        self.name = None
        self._maxVal = 0
        self._minVal = 0



from bridges.input_helper import *
from bridges.gamebase import *
import queue

class BlockingGame(KeyPressListener, GameBase):

    def __init__(self, assid, login, apikey, cols, rows):
        super(BlockingGame, self).__init__(assid, login, apikey, cols, rows)
        self.blocking_init()

    def blocking_init(self):
        self.q = queue.Queue(maxsize = 9)

    def key_press(self, keypress):
        try:
            type = str(keypress.get("key"))
            key = str(keypress.get("key"))
        except ValueError as e:
            traceback.print_tb(e.__traceback__)
            return

        if type == "keydown":
            self.q.put(key)

    def get_keypress(self):
        return self.q.get()

    def start(self):
        self.initialize()
        game_started = True

        while game_started:
            self.game_loop()
            self.render()
